#!/usr/bin/perl -w

# box version 0.9.99 - Copyright (C) Grzegorz Dbrowski 'kaliber'
# This may be freely redistributed under the terms of the GNU Public License.
# $Id: box.pl,v 1.167 2006/01/23 19:38:47 kaliber Exp $

# TODO
# - export exit code from script
# - implement -s -d (debug when building source package)
# - implement -sb (build binary from XML BOX - source & build)
# - implement -tidy option
# - implement -u option (uninstall package)
# - implement multipackage system (not now, after pingwinek-1.0 propably)

# - all sources exists
# - all patches exists
# - checking sources (archives eg. gzip, bzip2, zip) for errors
# - checking sources md5
# - patches apply corectly?
# - root for main source is good?
# - URL is valid (regex)?
# --------------

# Package naming:
# name-type-version-release.architecture.box
#   'type' is eg. devel or static or nothing
#   'architecuture' is eg. i386, athlon, src, noarch

use XML::XPath;
use Net::FTP;
use Date::Calc qw(Delta_Days);
use Date::Format;
use LWP::UserAgent;

use strict;

# Global variables
my %config;
my $TEMPDIR;
my $CVSBOXES;
my $BOXES;
my $LIBDIR;
my $DISTFILES;
my $DISTFILES_EXTRA;
my $PATCHES;
my $COMPRESS;
my $LOGDIR;
my $TESTSDIR;
my $BOXROOT;

sub xpath
{
  my $file = $_[0];
  my $xpstring = $_[1];
  my $xpath;
  my @result = ();
  if ("$file" eq '-') {
    $xpath = XML::XPath->new(ioref => \*STDIN);
  } else {
    $xpath = XML::XPath->new(filename => "$file");
  }
  my $nodes = $xpath->findnodes($xpstring);
  foreach my $node ($nodes->get_nodelist)
  {
    my $str = $node->string_value;
    push @result, "$str";
  }
  return @result;
}

sub getfirst
{
  my $res;
  if (scalar(@_) > 0) { $res = $_[0]; } else { $res = ""; }
  return $res;
}

sub basename
{
  my @array = split(/\//, $_[0]);
  return $array[$#array];
}

sub dirname
{
  my @array = split(/\//, $_[0]);
  my $result = '';
  for (my $i = 0; $i < $#array; $i++) {
    $result .= $array[$i] . "/";
  }
  return $result;
}

sub format_time
{
  my (@starttime) = (@_);
  my $year = $starttime[5]+1900;
  my $month = sprintf("%0.2d",$starttime[4] + 1);
  my $day = sprintf("%0.2d",$starttime[3]);
  my $hour = sprintf("%0.2d",$starttime[2]);
  my $min = sprintf("%0.2d",$starttime[1]);
  my $second = sprintf("%0.2d",$starttime[0]);
  return "$year-$month-$day-$hour:$min:$second";
}

sub macro
{
  my $str = $_[0];
  my $ver = $_[1];
  my $main_version = "";
  my @numbers = split('\.', $ver);
  my @numbers2 = split('\-', $ver);
  # main_version(2.6.11-rc1) = 2.6
  # main_version(6.1.9-1) = 6.1.9
  # main_version(2.4.2) = 2.4

  # we have '-' in version
  if (scalar(@numbers2) >= 2)
  {
    if ($numbers2[1] =~ /^[0-9]+$/) {
      $main_version = "$numbers2[0]";
    } elsif (scalar(@numbers) >= 2) {
      $main_version = "$numbers[0].$numbers[1]";
    } else {
      $main_version = $ver;
    }
  } else
  {
    if (scalar(@numbers) >= 2) {
      $main_version = "$numbers[0].$numbers[1]";
    } else {
      $main_version = $ver;
    }  
  }
  if ((scalar(@numbers2) == 1) && (scalar(@numbers) >= 2))
  {
    $main_version = "$numbers[0].$numbers[1]";
  }  
  $str =~ s/\%\{version\}/$_[1]/g;
  $str =~ s/\$VERSION/$_[1]/g;
  $str =~ s/\%\{main_version\}/$main_version/g;
  $str =~ s/\%\{GNOME.ORG\}/http\:\/\/ftp.gnome.org\/pub\/GNOME\/sources\//g;
  $str =~ s/\%\{GNU.ORG\}/ftp\:\/\/ftp.gnu.org\/gnu\//g;
  $str =~ s/\%\{SF.NET\}/http\:\/\/dl.sourceforge.net\/sourceforge/g;
  $str =~ s/\%\{CPAN.ORG\}/http\:\/\/www.cpan.org\/modules\/by\-module\//g;
  $str =~ s/\%\{make\}/
echo \"\* Detecting build requires (pkg\-config)\"
if \[ \-f pkg\-config\.txt \]\; then pkg\-config\-anal \> pkg\-config\-anal\.txt; fi
if \[ \-f configure \]\; then \.\/configure --help \> configurehelp.txt; fi
make/gx;
  $str =~ s/\%\{make_ko\}/
echo \"\* Detecting build requires (pkg\-config)\"
if \[ \-f pkg\-config\.txt \]\; then pkg\-config\-anal \> pkg\-config\-anal\.txt; fi
if \[ \-f configure \]\; then \.\/configure --help \> configurehelp.txt; fi;
KDIR=\/lib\/modules\/\$LINUX\/build; make KDIR=\$KDIR/gx;
  $str =~ s/\%\{make_old\}/make \$MAKE/g;
  $str =~ s/\%\{make_install\}/make install DESTDIR=\$ROOT_BIN GCONF_DISABLE_MAKEFILE_SCHEMA_INSTALL=1 MKINSTALLDIRS=\/usr\/share\/automake\/mkinstalldirs/g;
  $str =~ s/\%\{make_install_ko\}/make install DESTDIR=\$ROOT_BIN GCONF_DISABLE_MAKEFILE_SCHEMA_INSTALL=1 MKINSTALLDIRS=\/usr\/share\/automake\/mkinstalldirs INSTALL_MOD_PATH=\$ROOT_BIN MODPREFIX=\$ROOT_BIN KDIR=\$KDIR/g;
  $str =~ s/\%\{make_install_old\}/make install \$MAKEINSTALL/g;
  $str =~ s/\%\{python_setup\}/python setup.py/g;
  $str =~ s/\%\{python_build\}/python setup.py build/g;
  $str =~ s/\%\{python_install\}/python setup.py install --root=\$ROOT_BIN --prefix=\$PREFIX/g;
  $str =~ s/\%\{perl_make\}/perl Makefile.PL/g;
  $str =~ s/\%\{configure\}/\.\/configure --x-includes=\/usr\/include --x-libraries=\/usr\/lib --build=i686-pc-linux-gnu  \\\n\tPKG_CONFIG=pkg-config\\\n\tPATH=\$PWD:\$PATH\\\n\t\--disable-static\\\n\t\$CONFIG/g;
  $str =~ s/\%\{configure_old\}/\.\/configure --x-includes=\/usr\/include --x-libraries=\/usr\/lib  --build=i686-pc-linux-gnu\\\n\t\--disable-static\\\n\t\$CONFIG /g;
  $str =~ s/\%\{autoconf\}/autoconf/g;
  $str =~ s/\%\{automake\}/automake \-a \-c \-f/g;
  $str =~ s/\%\{aclocal\}/aclocal/g;
  $str =~ s/\%\{autoheader\}/autoheader/g;
  $str =~ s/\%\{gettextize\}/glib\-gettextize \-c \-f/g;
  $str =~ s/\%\{libtoolize\}/libtoolize \-c \-f/g;
  $str =~ s/\%\{intltoolize\}/intltoolize \-c \-f/g;
  $str =~ s/\%\{autoreconf\}/autoreconf \-vif/g;
  $str =~ s/\%\{desktopdir\}/\$DATADIR\/applications/g;
  $str =~ s/\%\{pixmapsdir\}/\$DATADIR\/pixmaps/g;
  $str =~ s/\%\{docdir\}/\$DATADIR\/doc/g;
  $str =~ s/\%\{ROOT_BIN\}/\$ROOT_BIN/g;
  $str =~ s/\%\{ROOT_DEV\}/\$ROOT_DEV/g;
  $str =~ s/\%\{ROOT_STA\}/\$ROOT_STA/g;
  $str =~ s/\%\{prefix\}/\$PREFIX/g;
  $str =~ s/\%\{execprefix\}/\$EXECPREFIX/g;
  $str =~ s/\%\{bindir\}/\$BINDIR/g;
  $str =~ s/\%\{sbindir\}/\$SBINDIR/g;
  $str =~ s/\%\{libexecdir\}/\$LIBEXECDIR/g;
  $str =~ s/\%\{datadir\}/\$DATADIR/g;
  $str =~ s/\%\{sysconfdir\}/\$SYSCONFDIR/g;
  $str =~ s/\%\{localstatedir\}/\$LOCALESTATEDIR/g;
  $str =~ s/\%\{libdir\}/\$LIBDIR/g;
  $str =~ s/\%\{infodir\}/\$INFODIR/g;
  $str =~ s/\%\{mandir\}/\$MANDIR/g;
  $str =~ s/\%\{includedir\}/\$INCLUDEDIR/g;
  $str =~ s/\%\{localedir\}/\$LOCALEDIR/g;
  $str =~ s/\%\{pkgconfigdir\}/\$LIBDIR\/pkgconfig/g;
  $str =~ s/\%\{aclocaldir\}/\$DATADIR\/aclocal/g;
  $str =~ s/\%\{javadir\}/\$LIBDIR\/java/g;
  $str =~ s/\%\{jredir\}/\$LIBDIR\/java\/jre/g;
  $str =~ s/\%\{javaclassdir\}/\$DATADIR\/java/g;
  $str =~ s/\%\{man1dir\}/\$MANDIR\/man1/g;
  $str =~ s/\%\{man2dir\}/\$MANDIR\/man2/g;
  $str =~ s/\%\{man3dir\}/\$MANDIR\/man3/g;
  $str =~ s/\%\{man4dir\}/\$MANDIR\/man4/g;
  $str =~ s/\%\{man5dir\}/\$MANDIR\/man5/g;
  $str =~ s/\%\{man6dir\}/\$MANDIR\/man6/g;
  $str =~ s/\%\{man7dir\}/\$MANDIR\/man7/g;
  $str =~ s/\%\{man8dir\}/\$MANDIR\/man8/g;
  $str =~ s/\%\{depmod\}/depmod \-a/g;
  $str =~ s/\%\{install\-fonts\}/fc\-cache \-v \-f/g;
  $str =~ s/\%\{scrollkeeper\-update\}/scrollkeeper\-update/g;
  $str =~ s/\%\{update\-mime\}/update\-mime\-database \/usr\/share\/mime
update\-desktop\-database
/gx;
  $str =~ s/\%\{clean_langs\}/function_clean_langs/gx;
  $str =~ s/\%\{clean_la\}/
find \$ROOT_BIN\$LIBDIR -mindepth 2 -name \'\*.la\' -exec rm -f \{\} \\;
/gx;
  $str =~ s/\%\{clean_all_la\}/
find \$ROOT_BIN\$LIBDIR -name \'\*.la\' -exec rm -f \{\} \\;
/gx;
  $str =~ s/\%\{categories\}/function_categories/gx;
  # obsolete
  $str =~ s/\$PROGRAM/$_[2]/g;
  return $str;
}

sub prefixname
{
  $_ = $_[0];
  s/\.tar\.gz$//g;
  s/\.tar\.bz2$//g;
  s/\.tgz$//g;
  s/\.tbz$//g;
  s/\.zip$//g;
  return $_;
}

# param1: dir
# param2: min logs
# param3: life
sub file_management
{
  my ($latestdir, $minlogs, $life) = @_;
  my $pattern = '^([0-9]{4})\-([0-1][0-9])\-([0-3][0-9])\-([0-2][0-9]):([0-5][0-9]):([0-5][0-9])\-.+';
  opendir(DIR,$latestdir);
  my @dirs = sort grep{ /$pattern/ && -d "$latestdir/$_" } readdir(DIR);
  my $nlogs = scalar(@dirs);
  my $i = 0;
  foreach (@dirs) {
    if ($i < $nlogs - $minlogs) {
      my @localtime = localtime(time);
      my ($d1, $m1, $y1) = ($localtime[3], $localtime[4]+1, $localtime[5]+1900);
      /$pattern/;
      my ($d2, $m2, $y2) = ($3, $2, $1);
      my $days = Delta_Days($y2, $m2, $d2, $y1, $m1, $d1);
      if ($days >= $life) {
	system "rm -f $latestdir/$_/*.log*";
	system "rmdir $latestdir/$_ &>/dev/null";
      }
    }
    $i++;
  }
  closedir(DIR);
}

sub logs_management
{
  # remove files older then 7 days if count(logs) > 200
  file_management("$LOGDIR/trunk/", 200, 7);
  # remove files older then 7 days if count(logs) > 500
  file_management("$TESTSDIR/trunk/", 500, 7);
}

# ============================================================================
# usage
# ============================================================================

sub usage()
{
  print q{
box version 0.9.99 - Copyright (C) Grzegorz Dbrowski 'kaliber'
This may be freely redistributed under the terms of the GNU Public License.

Usage: box [OPTION] [...]

  -s [-d] [-r tag] xml_package       Generate source package from package.box
      -r tag                         Use specific tag from CVS
      -d                             Debug (does not remove build script)
  -b source_package                  Build package from source package
  -t [-r tag] xml_package            Executing tests (sources, patches, etc.)
  -l bin_package                     List files in package and display info
  -i [-root] bin_package             Install package to system
      -root root_dir                 Use alternate dir instead of "/"
  -g [-f] [-r tag] xml_package       Download sources needed by package.box
      -f                             Force overwriting files
      -r tag                         Use specific tag from CVS
  -ga [-f] [-r tag]                  Download sources needed by all packages
      -f                             Force overwriting
      -r tag                         Use specific tag from CVS
  -todo [-r tag]                     Print TODO list for all packages
      -r tag                         Use specific tag from CVS
  -diff bin_package bin_package      Shows files difference in binary packages
  -validate xml_package              Validates XML package
  -init                              Creates directory hierarchy to work
  -template [URL]                    Print template box using URL to source
  -check  source_package             Ckecks for difference betwean packages
                                     and packages in the factory
  -accept source_package             Moves packages to the factory and backup
                                     old packages to the factory-old
  -checkdep groups                   Check dependency consistence in groups
  -checkbuilddep groups              Check build dependency consistence
  -checkfiles groups                 Check file conflicts in the factory
  -tidy xml_package                  Tidy xml package (eg. trailling spaces)
  -report                            Generates report about the packages
  -newsoft                           Check for new software (require internet)
  -proposelog diff_xml_package       Propose ChangeLog using diff file
  -distfiles                         Remove unused files from DISTFILES
  -cvs2db                            Update database about sources from CVS
  -prepatch source_package           Prepare data for patches in temp directory
  -bsdiff bin_package1 bin_package2  Creates binary patch
  -bspatch bin_package patch         Apply binary patch
  -show cache_file bin_package       Show file from cache (e.g. requires)
  -deps package_name                 Show build and install package dependencies
  -deps_build package_name           Show build package dependencies
  -deps_install package_name         Show install package dependencies
  -what_provide name                 Show package which provide name
};
  exit;
}

# ============================================================================
# read_config
# ============================================================================

sub read_config
{
  open(CONFIG,"/etc/box.conf") or die "Error: configuration file not found!";
  while (<CONFIG>)
  {
    chomp;
    my $line = $_;
    $line =~ /([^\s]*)[\s]*=[\s]*([^\s]*)/;
    $config{$1}=$2;
  }
  close(CONFIG);
  $TEMPDIR = $config{"TEMPDIR"};
  $CVSBOXES = $config{"CVSBOXES"};
  $BOXES = $config{"BOXES"};
  $LIBDIR = $config{"LIBDIR"};
  $DISTFILES = $config{"DISTFILES"};
  $DISTFILES_EXTRA = $config{"DISTFILES_EXTRA"};
  $PATCHES = $config{"PATCHES"};
  $LOGDIR = $config{"LOGDIR"};
  $TESTSDIR = $config{"TESTSDIR"};
  $BOXROOT = $config{"BOXROOT"};

  $COMPRESS="";
}

# ============================================================================
# download_source(package.box, force?, tag?)
# ============================================================================
# FIXME: cleaning temporary files before die!
sub download_source
{
  my $file = $_[0];
  my $force = $_[1];
  my $tag = $_[2];
  my @temp = ();
  my $V;
  my $P;
  my @URLS = ();
#  print "File: $file\n";
  if ($tag ne '0') {
    print "Tag defined: :$tag:\n";
  }
#  if ($force == 1) {
#    print "Force overwriting\n";
#  }
  if ($tag ne '0') {
    system "mkdir -p $TEMPDIR; cd $TEMPDIR; cvs co -r $tag $CVSBOXES/$file";
    $file = "$TEMPDIR/$CVSBOXES/$file";
  } else
  {
    if (index($file,'/') == -1)
    {
      $file = "$BOXES/$file";
    }
  }
  if (! -f $file) {
    die "File $file not found!\n";
  }
  @temp = xpath("$file",'/box/source/name/@version');
  $V = getfirst(@temp);
  @temp = xpath("$file",'/box/source/name');
  $P = getfirst(@temp);
  @temp = xpath("$file",'/box/source/url');
  foreach my $url (@temp)
  {
    $url = macro($url, $V, $P);
    if (length($url) > 0) {
      push @URLS, $url;
    }
  }
  foreach my $URL (@URLS)
  {
    my $filename = basename($URL);
    if (-f "$DISTFILES/$filename") {
      if ($force eq '1') {
        print STDERR "Forceing: deleting $DISTFILES/$filename\n";
        system "rm -f $DISTFILES/$filename";
        system "cd $DISTFILES; wget -N --passive-ftp -t 3 $URL";
      }
    } else
    {
      system "cd $DISTFILES; wget -N --passive-ftp -t 3 $URL";
    }
  }
  if ($tag ne '0') {
    system "rm $file";
    system "rm -rf $TEMPDIR/$CVSBOXES/CVS";
    system "rmdir $TEMPDIR/$CVSBOXES";
  }
}

# ============================================================================
# download_all_sources(force?, tag?)
# ============================================================================
# FIXME: cleaning temporary files before die!
sub download_all_sources
{
  my $allfiles;
  my $force = $_[0];
  my $tag = $_[1];
  my @temp = ();
  my $V;
  my $P;
  my @URLS = ();
  my @files = ();
#  print "File: $file\n";
  if ($tag ne '0') {
    print "Tag defined: :$tag:\n";
  }
#  if ($force == 1) {
#    print "Force overwriting\n";
#  }
  if ($tag ne '0') {
    system "mkdir -p $TEMPDIR; cd $TEMPDIR; cvs co -r $tag $CVSBOXES";
    $allfiles = "$TEMPDIR/$CVSBOXES/*.box";
  } else
  {
    $allfiles = "$BOXES/*.box";
  }
  @files = glob($allfiles);
  foreach my $file (@files)
  {
    @temp = xpath("$file",'/box/source/name/@version');
    $V = getfirst(@temp);
    @temp = xpath("$file",'/box/source/name');
    $P = getfirst(@temp);
    @temp = xpath("$file",'/box/source/url');
    print "Checking $P...\n";
    foreach my $url (@temp)
    {
      $url = macro($url, $V, $P);
      if (length($url) > 0) {
        push @URLS, $url;
      }
    }
    foreach my $URL (@URLS)
    {
      my $filename = basename($URL);
      if (-f "$DISTFILES/$filename") {
        if ($force eq '1') {
          print STDERR "Forceing: deleting $DISTFILES/$filename\n";
          system "rm -f $DISTFILES/$filename";
          system "cd $DISTFILES; wget -N --passive-ftp -t 3 $URL";
        }
      } else
      {
        system "cd $DISTFILES; wget -N --passive-ftp -t 3 $URL";
      }
    }
    @URLS = ();
    @temp = ();
  }
  if ($tag ne '0') {
    system "rm -rf $TEMPDIR/$CVSBOXES";
  }
}


# ============================================================================
# generate_source_package(package.box, tag, debug)
# ============================================================================

# FIXME: cleaning temporary files before die!
sub generate_source_package
{
  my $file = $_[0];
  my $origfile = $file;
  my $tag = $_[1];
  my $debug = $_[2];
  my @temp = ();
  my $V;
  my $P;
  my $R;
  my @PATCHES = ();
  my @URLS = ();

  print "Generating source package for $file...\n";
#  if ($tag ne '0') {
#    print "Tag defined: $tag\n";
#  }
  if ($tag ne '0') {
    system "mkdir -p $TEMPDIR; cd $TEMPDIR; cvs co -r $tag $CVSBOXES/$file";
    $file = "$TEMPDIR/$CVSBOXES/$file";
  } else
  {
    if (index($file,'/') == -1)
    {
      $file = "$BOXES/$file";
    }
  }
  if (! -f $file) {
    die "* File $file not found!\n";
  }
  # reading some information from package.box
  # get name, get version, release,
  @temp = xpath("$file",'/box/source/name/@version');
  $V = getfirst(@temp);
  @temp = xpath("$file",'/box/source/name/@release');
  $R = getfirst(@temp);
  @temp = xpath("$file",'/box/source/name');
  $P = getfirst(@temp);
  @temp = xpath("$file",'/box/source/patch');
  foreach my $patch (@temp)
  {
    $patch = macro($patch, $V, $P);
    if (length($patch) > 0) {
      push @PATCHES, $patch;
    }
  }
  @temp = xpath("$file",'/box/source/url');
  foreach my $url (@temp)
  {
    $url = macro($url, $V, $P);
    if (length($url) > 0) {
      push @URLS, $url;
    }
  }
  use File::Temp qw/ tempfile tempdir /;
  my $tempdir = tempdir ( "$TEMPDIR/package.XXXX" );
  # preparing directory
  system "mkdir -p $tempdir/sources";
  system "mkdir -p $tempdir/patches";
  system "cp $file $tempdir/package.box";
  # copying sources and patche here
  foreach my $patch (@PATCHES)
  {
    # FIXME: patches are get from directory now and should be get from cvs
    if (! -f "$PATCHES/$patch") {
      die "* Patch $patch not found!";
    }
    system "cp $PATCHES/$patch $tempdir/patches";
  }
  foreach my $url (@URLS)
  {
    my $filename = basename($url);
    my $realfile = "$DISTFILES/$filename";
    if (! -f "$DISTFILES/$filename") {
      print "* File $filename not found. Trying to get it :)\n";
      if ($url =~ /http:\/\/|ftp:\/\//) {
        download_source("$file", "0", "$tag");
      }
      if (! -f "$DISTFILES/$filename") {
        print "* Searching in DISTFILES_EXTRA\n";
        if (-f "$DISTFILES_EXTRA/$filename") {
            print "* Found.\n";
            $realfile = "$DISTFILES_EXTRA/$filename";
        } else
        {
          system "rm -rf $tempdir";
          die "* File $filename not found!\n";
        }
      }
    }
    system "cp -f $realfile $tempdir/sources";
  }
  chomp(my $cwd = `pwd`);
  system "cd $tempdir; tar cf$COMPRESS $cwd/$P-$V-$R.src.box *";
  # cleaning
  if ($tag ne '0') {
    system "rm -rf $TEMPDIR/$CVSBOXES/CVS";
    system "rm $file";
    system "rmdir $TEMPDIR/$CVSBOXES";
  }
  system "rm -rf $tempdir";
}

# ============================================================================
# build_package
# ============================================================================

sub build_package
{
  my $file = $_[0];
  my @temp = ();
  my $V;
  my $P;
  my $R;
  my $fails;

  my $FILE = "package.box";
  open(CMD,"tar xf$COMPRESS"." $file $FILE|");
  close(CMD);
  @temp = xpath("$FILE",'/box/source/name/@version');
  $V = getfirst(@temp);
  @temp = xpath("$FILE",'/box/source/name/@release');
  $R = getfirst(@temp);
  @temp = xpath("$FILE",'/box/source/name');
  $P = getfirst(@temp);
  $fails = getfirst(xpath("$FILE",'/box/build/fails'));

  open(OUT,">$P.sh");
  ### Generatin script for build package ###
  print OUT "#!/bin/bash\n";
  print OUT "# This file is generated please do not edit!\n";
  print OUT '# $Id: box.pl,v 1.167 2006/01/23 19:38:47 kaliber Exp $',"\n";
  print OUT "# file: $file\n\n";

  print OUT "umask 022\n";
  print OUT "FAILS=\"$fails\"\n";
  print OUT q{
if [ "x$FAILS" != "x" ]
then
  echo "===================WARNING============================="
  echo "Building will fails:"
  echo "$FAILS"
  echo "======================================================="
fi
};

  print OUT "# Setting default variables\n";
  print OUT "TMP=\"$TEMPDIR\"\n";
  print OUT "PACKAGESRC=\"$file\"\n";
  print OUT "STARTBUILD=\$1\n";
  print OUT q{
LC_ALL=POSIX; export LC_ALL
LANG=POSIX; export LANG
LINGUAS=pl; export LINGUAS
WX_LINGUAS=pl; export WX_LINGUAS
};
  print OUT "TMPDIR=\$TMP/$P-$V-$R-\$\$";
  print OUT q{
STARTDIR=$PWD
SRCROOT=$TMPDIR/src
};
  print OUT "PROGRAM=$P\n";
  print OUT "VERSION=$V\n";
  print OUT 'PACKAGENAME=$PROGRAM',"\n";
  print OUT "RELEASE=$R\n";
  print OUT 'ROOT_BIN=$TMPDIR/ROOT_BIN',"\n";
  print OUT 'ROOT_DEV=$TMPDIR/ROOT_DEV',"\n";
  print OUT 'ROOT_STA=$TMPDIR/ROOT_STA',"\n";

  print OUT q{
PREFIX="/usr"
};
  @temp = xpath("$FILE",'/box/define/@name');
  foreach my $define_name (@temp)
  {
    my @define_values = xpath("$FILE","/box/define[\@name = \"$define_name\"]");
    my $define_value = getfirst(@define_values);
    print OUT $define_name, '="', $define_value,'"',"\n";
  }
  print OUT q{
EXECPREFIX="$PREFIX"
BINDIR="$EXECPREFIX/bin"
SBINDIR="$EXECPREFIX/sbin"
LIBEXECDIR="$EXECPREFIX/libexec"
DATADIR="$PREFIX/share"
SYSCONFDIR="/etc"
SHAREDSTATEDIR="$PREFIX/com"
LOCALSTATEDIR="/var"
LIBDIR="$EXECPREFIX/lib"
INFODIR="/usr/share/info"
MANDIR="/usr/share/man"
INCLUDEDIR="$PREFIX/include"
OLDINCLUDEDIR="/usr/include"
LOCALEDIR="/usr/share/locale"
};
  print OUT "\n";
  print OUT q{CONFIG="--prefix=$PREFIX \
	--exec-prefix=$EXECPREFIX \
	--bindir=$BINDIR \
	--sbindir=$SBINDIR \
	--libexecdir=$LIBEXECDIR \
	--datadir=$DATADIR \
	--sysconfdir=$SYSCONFDIR \
	--sharedstatedir=$SHAREDSTATEDIR \
	--localstatedir=$LOCALSTATEDIR \
	--libdir=$LIBDIR \
	--infodir=$INFODIR \
	--mandir=$MANDIR \
	--includedir=$INCLUDEDIR \
	--oldincludedir=$OLDINCLUDEDIR"

};

  print OUT q{MAKE="prefix=$PREFIX \
	exec-prefix=$EXECPREFIX \
	bindir=$BINDIR \
	sbindir=$SBINDIR \
	libexecdir=$LIBEXECDIR \
	datadir=$DATADIR \
	sysconfdir=$SYSCONFDIR \
	sharedstatedir=$SHAREDSTATEDIR \
	localstatedir=$LOCALSTATEDIR \
	libdir=$LIBDIR \
	infodir=$INFODIR \
	mandir=$MANDIR \
	includedir=$INCLUDEDIR \
	oldincludedir=$OLDINCLUDEDIR"

};

  print OUT q{MAKEINSTALL="prefix=$ROOT_BIN$PREFIX \
	exec-prefix=$ROOT_BIN$EXECPREFIX \
	bindir=$ROOT_BIN$BINDIR \
	sbindir=$ROOT_BIN$SBINDIR \
	libexecdir=$ROOT_BIN$LIBEXECDIR \
	datadir=$ROOT_BIN$DATADIR \
	sysconfdir=$ROOT_BIN$SYSCONFDIR \
	sharedstatedir=$ROOT_BIN$SHAREDSTATEDIR \
	localstatedir=$ROOT_BIN$LOCALSTATEDIR \
	libdir=$ROOT_BIN$LIBDIR \
	infodir=$ROOT_BIN$INFODIR \
	mandir=$ROOT_BIN$MANDIR \
	includedir=$ROOT_BIN$INCLUDEDIR \
	oldincludedir=$ROOT_BIN$OLDINCLUDEDIR"

};

  print OUT q{
function CLEAN
{
  #echo "* rm -rf $TMPDIR"
  echo "* Removing temporary files..."
  rm -rf $TMPDIR
  rm -f $STARTDIR/$PROGRAM.sh
}

function STRIP() # directory
{
  echo -n "* Optimizing (stripping binaries): "
  for f in `find $1 -type f \( -perm -0100 -or -perm -0010 -or -perm -0001 \) -exec file {} \; | \
	grep -v ' shared object,' | \
	sed -n -e 's/^\(.*\):[ 	]*ELF.*, not stripped/\1/p'`; do
	strip $f || :
  done

  for f in `find $1 -type f -a -exec file {} \; | grep ' shared object,' | \
	  sed -n -e  's/^\(.*\):[  ]*ELF.*, not stripped/\1/p'`; do
	  strip --strip-unneeded $f
  done

}
};
  print OUT q{
function GZIPDOC() # directory
{
  echo -n "* Optimizing (compressing docs)  : "
  cd $1
  OLD=$PWD
  SEARCH=""
  if [ ! -z $MANDIR ]; then
   if [ -d .$MANDIR  ]; then SEARCH="$SEARCH .$MANDIR"; fi
  fi
  if [ ! -z $INFODIR ]; then
   if [ -d .$INFODIR ]; then SEARCH="$SEARCH .$INFODIR"; fi
  fi
  # delete file dir and INFO
  rm -f $1/$INFODIR/dir
  # delete unused perlocal.pod (maybe do this in other place?)
  rm -f $1/$LIBDIR/perl5/*/*/perllocal.pod
  find $1/$LIBDIR/perl5/site_perl -name '.packlist' -exec rm -f {} \; 2>/dev/null
  find $1/$LIBDIR/perl5/vendor_perl -name '.packlist' -exec rm -f {} \; 2>/dev/null

  if [ ! -z "$SEARCH" ]
  then
    # find all normal not compressed files
    for f in `find $SEARCH -type f ! -regex '.*\.gz'`
    do
      [ -f $f ] || continue
      n=`find $f -printf "%n\n"` # 1 -> is not hardlink
      if [ $n -eq 1 ]; then gzip $f;
      else
	# special for hardlinks
	inode=`find $f -printf '%i'`
	others=`find $SEARCH -type f -inum $inode`
	for afile in $others; do [ "$afile" != "$f" ] && rm -f $afile; done
	gzip $f
	for afile in $others; do [ "$afile" != "$f" ] && ln $f.gz $afile.gz; done
      fi
    done
    # special for symlinks
    TMP_SYMLINK="$TMPDIR/box.$$"
    find $SEARCH -type l > $TMP_SYMLINK
    while read f
    do
      l=`find "$f" -printf '%l'`
      rm -f "$f"
      ln -sf "${l%.gz}.gz" "${f%.gz}.gz"
    done < $TMP_SYMLINK
    rm -f "$TMP_SYMLINK"
  fi
  cd $OLD
}
};

  print OUT q{
function DESKTOP_FIX_LOCATION()
{
  if [ -d $ROOT_BIN/usr/share/control-center-2.0/capplets ]; then
    mkdir -p $ROOT_BIN/usr/share/applications/
    mv $ROOT_BIN/usr/share/control-center-2.0/capplets/*.desktop $ROOT_BIN/usr/share/applications/
  fi
}

function DESKTOP_CLEANER() # directory
{
  cd $1
  if [ -d ./usr/share/applications ]; then
    TEMPPROG="/tmp/tempprog1$$"
    cat << EOF > $TEMPPROG
#!/usr/bin/perl

my @langs = ("pl", "pl_PL");
my @bad_categories = ("Application", "X-Ximian-Main", "X-Red-Hat-Base");
my @valid_onlyshowin = ("KDE", "GNOME", "ROX", "XFCE", "Old");
my \$print;
my @desktop_lines;
my \$is_encoding = 0;

open(FILE, \$ARGV[0]);
while (<FILE>)
{
  chomp;
  # generic fixes
  s/(.*)\;\$/\$1/;
  s/^Nume(.*)/Name\$1/;
  s/^Genericname/GenericName/;
  s/^StartupNotification/StartupNotify/;
  ###
  my \$line = \$_;
  \$print = 0;
  if (/^(Name|Comment|GenericName)\[(.+)\]\=(.*)\$/) {
    \$desktop_lang = \$2;
    \$print = 0;
    foreach my \$lang (@langs) {
      if (\$lang eq \$desktop_lang) {
        \$print = 1;
      }
    }
  } elsif (/^Icon=(.+)/) {
    my \$icon = \$1;
    if (\$icon =~ /^[^\/].+\/.+/) {
      \$line = "Icon=/usr/share/pixmaps/\$icon";
    }
    \$print = 1;
  } elsif (/^X\-Desktop\-File\-Install\-Version=(.*)\$/) {
    # remove this line
  } elsif (/^Encoding=(.*)\$/) {
    if (\$1 eq "Legacy-Mixed") {
      \$line="Encoding=UTF-8";
    }
    \$print = 1;
    \$is_encoding = 1;
  } elsif (/^Categories=(.+)\$/) {
    my @categories = split(/\;/, \$1);
    \$line = "Categories=";
    foreach my \$category (@categories) {
      my \$exists = 0;
      foreach \$bad_category (@bad_categories) {
        if (\$category eq \$bad_category) {
	  \$exists = 1;
	}
      }
      if (\$exists == 0) {
        \$line .= "\$category;";
      }
    }
    \$print = 1;
  } elsif (/^DocPath=(.*)\$/) {
    \$line = "X-GNOME-DocPath=\$1";
    \$print = 1;
  } elsif (/^OnlyShowIn=(.*)\$/) {
    my @elements = split(/\;/, \$1);
    \$line = "OnlyShowIn=";
    foreach my \$element (@elements) {
      \$element =~ s/XFCE4/XFCE/g;
      \$exists = 0;
      foreach my \$valid_onlyshowin (@valid_onlyshowin) {
        if (\$element eq \$valid_onlyshowin) {
	  \$exists = 1;
	}
      }
      if (\$exists == 1) {
        \$line .= "\$element;";
     }
    }
    \$print = 1;
  } elsif (/^(Actions|MimeType)=(.*)\$/) {
    my @elements = split(/\;/, \$2);
    \$line = "\$1=";
    foreach my \$element (@elements) {
        \$line .= "\$element;";
     }
    \$print = 1;
  } elsif (/^Terminal=(.+)\$/) {
    my \$orig_line = \$line;
    if (\$1 eq "0") {
      \$line = "Terminal=false";
    } elsif (\$1 eq "1") {
      \$line = "Terminal=true";
    } else {
      \$line = \$orig_line;
    }
    \$print = 1;
  }
 else {
    \$print = 1;
  }
  if (\$print == 1) {
    #print "\$line\n";
    push @desktop_lines, \$line;
  }
}
close(FILE);

if (\$is_encoding == 0) {
  push @desktop_lines, "Encoding=UTF-8";
}

foreach (@desktop_lines) {
  print "\$_\n";
}
EOF
    chmod 755 $TEMPPROG
    for i in `find "$1/usr/share/applications" -type f`
    do
      mv $i $i.bak
      $TEMPPROG $i.bak > $i;
      rm -f $i.bak
    done
    rm -f "$TEMPPROG"
  fi
}

function OPTIMIZEGLADE() # directory
{
  cd $1
  echo -n "* Optimizing (glade)              : "
  for i in `find ./usr -type f -name '*.glade*'`
  do
    gzip -9 $i
    mv $i.gz $i
  done
}

function STRIPSCHEMAS() # directory
{
  cd $1
  echo -n "* Optimizing (stripping schemas) : "
  if [ -d ./etc/gconf/schemas ]; then
    TEMPPROG="/tmp/tempprog$$"
    cat << EOF > $TEMPPROG
#!/usr/bin/perl

\$TEMPFILE="/tmp/schemas\$\$";
\$filename=\$ARGV[0];
#print "Stripping: \$filename\n";

open(FILE,"\$filename");
while (<FILE>)
{
  if (/locale\ name\=\"(.*?)\"/)
  {
    \$_=\$1;
    s/@/\\@/g;
    \$language=\$_;

    if ("\$language" ne "pl" && "\$language" ne "C")
    {
      \$hash{"\$language"}=1;
    }
  }
}
@mykeys = keys %hash;
foreach (@mykeys)
{
  \$LNG=\$LNG."\$_|";
}
close(FILE);

open(FILE,"\$filename");
open(TMP,">\$TEMPFILE");
while (<FILE>)
{
    chomp;
    print TMP \$_;
}
close(TMP);
close(FILE);

open(TMP,"\$TEMPFILE");
open(FILE,">\$filename");
while (<TMP>)
{
  s|[ ]+|\ |g;		# squize spaces
  s|<!--.*?-->||g;	# strip coments
  s|>[ \t]+<|><|g;	# strip space betwean tags
  s|<locale\s+name=\"(\$LNG)\">.*?locale>||g;	# remove unused locale!
  print FILE "\$_\n";
}
close(FILE);
close(TMP);

system "rm -f \$TEMPFILE";
EOF
    chmod 755 $TEMPPROG
#    echo "* stripping schemas ($1)..."
    find "$1/etc/gconf/schemas" -type f -exec $TEMPPROG {} \;
    rm -f "$TEMPPROG"
  fi
}

function function_clean_langs()
{
[ -z $1 ] && return
echo "Clean rule: $1"
DELETE_LANGS=''
ALL_LANGS="C am ar af az be bg bn bs ca cs cy da de el en en_CA en_GB es es_ES es_MX et eu fa fi fr fr_CA ga gl gu gr he hi hr hu id is it it_IT ja ko lt lv mk ml mn ms nb ne nl nn nb_NO no pa pl pt pt_BR ro ru rw sk sl sq sr sr@Latn sv ta th tr uk vi wa zh zh_CN zh_TW zh_HK"
VALID_LANGS="C en $LINGUAS "

for i in $ALL_LANGS
do
  was=0
  for j in $VALID_LANGS
  do
    if [ $i = $j ]
    then
      was=1
    fi
  done
  if [ $was -eq 0 ]
  then
    DELETE_LANGS="$DELETE_LANGS $i"
  fi
done

  pattern=$ROOT_BIN/$1
  dir=`dirname $pattern`
  base=`basename $pattern`
  for i in `find $dir -maxdepth 0 2>/dev/null`
  do
    for j in $DELETE_LANGS
    do
      file=$(echo $i/$base | sed -e "s/%{lang}/$j/")
      for k in `find $file -maxdepth 0 2>/dev/null`
      do
        echo "- removing:" ${file#$ROOT_BIN/}
        rm -rf $file
      done
    done
  done
}

function function_clean_langs_default()
{
  function_clean_langs /usr/share/locale/%{lang}
  function_clean_langs /usr/share/man/%{lang}
  function_clean_langs /usr/share/gnome/help/*/%{lang}
  function_clean_langs /usr/share/omf/*/*-%{lang}.omf
}

function function_fix_symlinks()
{
cd $1
find . -type l > /tmp/fixlinks$$
while read line
do
  symlink=$(readlink $line)
  newlink=${symlink#$ROOT_BIN}
  newlink=${newlink#$ROOT_DEV}
  if [ "$symlink" != "$newlink" ]
  then
    ln -sf $newlink $line 
    echo "* Fixing symlink for $line"
  fi
done < /tmp/fixlinks$$
cd - >/dev/null
rm -f /tmp/fixlinks$$
}

function function_create_fhs()
{
  [ "$PROGRAM" == "pgw-base" ] && return
  mkdir -p $1/{bin,etc/{pam.d,rc.d/init.d,sysconfig},lib,sbin,srv,var/{lib,log},usr/{bin,include,lib,libexec,sbin,share/{doc,applications,application-registry,mime-info,info,pixmaps,man/man{1,2,3,4,5,6,7,8,n}}}}
}

function function_clean_fhs()
{
  [ "$PROGRAM" == "pgw-base" ] && return
  rmdir $1/{bin,etc/{pam.d,rc.d/init.d,sysconfig},lib,sbin,srv,var/{lib,log},var} &>/dev/null
  rmdir $1/etc/rc.d/init.d &>/dev/null
  rmdir $1/etc/rc.d &>/dev/null
  rmdir $1/etc &>/dev/null
  rmdir $1/usr/{bin,include,lib,libexec,sbin} &>/dev/null
  rmdir $1/usr/share/{doc,applications,application-registry,mime-info,info,pixmaps,man/man{1,2,3,4,5,6,7,8,n}} &>/dev/null
  rmdir $1/usr/share/man &>/dev/null
  rmdir $1/usr/share &>/dev/null
  rmdir $1/usr &>/dev/null
}


function function_categories()
{
  DESKTOP_FIX_LOCATION
  file=$1.desktop
  shift
  categories=$@
  for i in $categories
  do
    f=${i:0:1}
    echo "f: $file | $i";
    x=''
    if [ "x$f" == 'x+' ]
    then
      x=$(echo $i | perl -pe 's/\+(.+)/\-\-add\-category\=$1/')
    fi
    if [ "x$f" == 'x-' ]
    then
      x=$(echo $i | perl -pe 's/\-(.+)/\-\-remove\-category\=$1/')
    fi
    if [ "x$f" == 'x@' ]
    then
      x=$(echo $i | perl -pe 's/\@(.+)/\-\-add\-only\-show\-in\=$1/')
    fi
    cat="$cat $x "
  done
  orig="$ROOT_BIN$DATADIR/applications/$file"
  desktop-file-install \
	--vendor=pingwinek \
	--dir=$ROOT_BIN$DATADIR/applications \
	$cat $orig 2>/dev/null
  mv $ROOT_BIN$DATADIR/applications/pingwinek-$file $orig
}

function MAKEBOX() # directory
{
  PKGCONFIGANAL=$PWD/../builddir/pkg-config-anal.txt
  CONFIGUREHELP=$PWD/../builddir/configurehelp.txt
  NEWS=$PWD/../builddir/NEWS
  NEWS2=$PWD/../builddir/RELEASE-NOTES
  PERMS=$PWD/../builddir/perms.txt
  BUILDPROVIDES=$PWD/../builddir/buildprovides.txt
  PROVIDES=$PWD/../builddir/provides.txt
  REQUIRES=$PWD/../builddir/requires.txt
  BUILDDIR=$PWD/../builddir
  cd $1
  if [ -d "usr/local" ]
  then
    mkdir -p "usr"
    mv "usr/local" "."
    rmdir "usr"
    mv "local" "usr"
  fi
  if [ -d "usr/man" ]
  then
    mkdir -p "usr/share"
    mv "usr/man" "usr/share"
  fi
  if [ -d "usr/doc" ]
  then
    mkdir -p "usr/share"
    mv "usr/doc" "usr/share"
  fi
  # default permissions
  
  # workaround for pgw-base
  if [ "x$PROGRAM" != "xpgw-base" ]
  then
    chmod -Rf a+rX,g-w,o-w .
  fi
  chmod -f u+w .$BINDIR/* 2>/dev/null
  chmod -f u+w .$SBINDIR/* 2>/dev/null
  chmod -f u+w .$LIBDIR/* 2>/dev/null
  chmod -f u+w .$INCLUDEDIR/* 2>/dev/null
  ILE=`find $1 | wc -l`
  PACZKA=`basename $1`

  TYPE="$2" # 'devel' or 'static' or ''
  [ "x$TYPE" != "x" ] && TYPE="-$TYPE"
  ARCH="$3"
  FULLNAME="$PACKAGENAME$TYPE-$VERSION-$RELEASE.$ARCH"
#  echo "-------------------$TYPE--------------------------"
  if [ $ILE -ne 1 ]
  then
    echo "------------------------------------------------------------------------------"
    SIZE1=$(du -sb . | cut -f1)
    STRIP $1
    SIZE2=$(du -sb . | cut -f1)
    RATIO=$((100*SIZE2/SIZE1))
    echo "Actual size ratio $RATIO%"

    GZIPDOC $1
    SIZE2=$(du -sb . | cut -f1)
    RATIO=$((100*SIZE2/SIZE1))
    echo "Actual size ratio $RATIO%"

    STRIPSCHEMAS $1
    DESKTOP_CLEANER $1
    SIZE2=$(du -sb . | cut -f1)
    RATIO=$((100*SIZE2/SIZE1))
    echo "Actual size ratio $RATIO%"

    OPTIMIZEGLADE $1
    SIZE2=$(du -sb . | cut -f1)
    RATIO=$((100*SIZE2/SIZE1))
    echo "Actual size ratio $RATIO%"

#    echo "* Size after optimalization is $RATIO% of original size.";
#    echo "* mkdir -p $TMPDIR/$FULLNAME";
    # default permisions (after stripping)
#    [ -f ".$LIBDIR/lib*.so.*" ] && chmod -f 755 .$LIBDIR/lib*.so.*
    [ -d .$LIBDIR ] && find .$LIBDIR -maxdepth 1 -name "lib*.so.*" -exec chmod 755 {} \;
    
    mkdir -p $TMPDIR/$FULLNAME/data
    mkdir -p $TMPDIR/$FULLNAME/cache
#    echo "* PWD=$PWD; tar cfz $TMPDIR/$FULLNAME/$FULLNAME.tar.gz --posix"
    echo -n "* Writing $FULLNAME.box..."
    find etc/gconf/schemas/ -name *.schemas 2>/dev/null >/dev/null
    if [ $? == 0 ]; then
	echo "%{gconf_schemas_install}" >> $TMPDIR/$FULLNAME/cache/post-install
	echo "package.gconf" >> $REQUIRES
    fi
    [ -f "$PERMS" ] && cat $PERMS | sort -u -k 2 > $TMPDIR/$FULLNAME/cache/perms
    [ -f "$BUILDPROVIDES" ] && cat $BUILDPROVIDES | sort -u | perl -e 'print sort(<STDIN>);' > $TMPDIR/$FULLNAME/cache/buildprovides
    [ -f "$BUILDDIR/scrollkeeper-update.txt" ] && echo "package.scrollkeeper" >> $REQUIRES
    [ -d "usr/share/fonts" -o -f "usr/lib/X11/fonts" ] && echo "package.fontconfig" >> $REQUIRES
    [ -d "lib/modules" ] && echo "package.module-init-tools" >> $REQUIRES
    [ -f "$BUILDDIR/update-mime.txt" ] && echo "package.shared-mime-info" >> $REQUIRES
    [ -f "$BUILDDIR/update-mime.txt" ] && echo "package.desktop-file-utils" >> $REQUIRES
    [ -f "$PKGCONFIGANAL" ] && cp $PKGCONFIGANAL $TMPDIR/$FULLNAME/cache/buildrequires
    [ -f "$BUILDDIR/scrollkeeper-update.txt" ] && echo "%{scrollkeeper-update}" >> $TMPDIR/$FULLNAME/cache/post-install
    [ -d "lib/modules" ] && echo "%{depmod}" >> $TMPDIR/$FULLNAME/cache/post-install
    [ -d "lib/modules" ] && echo "package.module-init-tools" >> $REQUIRES
    [ -d "usr/share/fonts" -o -f "usr/lib/X11/fonts" ] && echo "%{install-fonts}" >> $TMPDIR/$FULLNAME/cache/post-install
#    [ -f "$BUILDDIR/depmod.txt" ] && echo "%{depmod}" >> $TMPDIR/$FULLNAME/cache/post-install
    [ -f "$BUILDDIR/update-mime.txt" ] && echo "%{update-mime}" >> $TMPDIR/$FULLNAME/cache/post-install
    [ -f "$BUILDDIR/ldconfig.txt" ] && echo "ldconfig" >> $TMPDIR/$FULLNAME/cache/post-install
    for i in `find -name fonts.dir -printf '%h\n'`
    do
	rm -f $i/fonts.dir
	i=${i#.}
	echo mkfontdir $i >> $TMPDIR/$FULLNAME/cache/post-install
	echo "package.mkfontdir" >> $REQUIRES
    done
    for i in `find -name fonts.scale -printf '%h\n'`
    do
	rm -f $i/fonts.scale
	i=${i#.}
	echo mkfontscale $i >> $TMPDIR/$FULLNAME/cache/post-install
	echo "package.mkfontscale" >> $REQUIRES
    done
    [ -f "$PROVIDES" ] && cat $PROVIDES | sort -u | perl -e 'print sort(<STDIN>);' > $TMPDIR/$FULLNAME/cache/provides
    [ -f "$REQUIRES" ] && cat $REQUIRES | sort -u | perl -e 'print sort(<STDIN>);' > $TMPDIR/$FULLNAME/cache/requires
    
    find * -type f > $TMPDIR/$FULLNAME/cache/files
    tar cfz $TMPDIR/$FULLNAME/data/$FULLNAME.tar.gz --owner=root --group=root --posix * 
    cd $SRCROOT
    # xml box should be in ant type of package
    tar xf $STARTDIR/$PACKAGESRC package.box
    cp package.box $TMPDIR/$FULLNAME
    # if ./configure exists then save help page
    if [ -f "$CONFIGUREHELP" ]
    then
      DIR=$(dirname "$CONFIGUREHELP")
      LINES=$(cat "$CONFIGUREHELP" | wc -l)
      cat "$CONFIGUREHELP" | tail -n $(( LINES - 7 )) > $TMPDIR/$FULLNAME/cache/configurehelp
      #cp "$CONFIGUREHELP" $TMPDIR/$FULLNAME/cache/configurehelp
    fi
    # if NEWS exists then copy it to cache
    if [ -f "$NEWS" ]
    then
      cat "$NEWS" | head -n 100 > $TMPDIR/$FULLNAME/cache/news
    else
      if [ -f "$NEWS2" ]
      then
        cat "$NEWS2" | head -n 100 > $TMPDIR/$FULLNAME/cache/news
      fi
    fi
    echo $STARTBUILD > $TMPDIR/$FULLNAME/cache/buildinfo
#    echo "SOURCE=$FILE" >> $TMPDIR/$NAME-$TYP/BOX/INFO
    cd $TMPDIR/$FULLNAME
    tar cf $STARTDIR/$FULLNAME.box  -b 1 *
    SIZE=$(du -k $STARTDIR/$FULLNAME.box | cut -f1)
    echo -n " ($SIZE"
    echo "kB)"
    IS_PACKAGE="1"
  else
#    echo "* no package $TYPE"
    IS_PACKAGE="0"
  fi
}
};
  print OUT q{
function DEVEL()
{
  cd $ROOT_DEV 
  N1=$(find . | wc -l)
  cd - >/dev/null 

  cd $ROOT_BIN
  cp -a --parent .$1 $ROOT_DEV
  rm -rf .$1

  cd $ROOT_DEV
  N2=$(find . | wc -l)
  cd - >/dev/null 

  if [ "x$N1" != "x$N2" ]
  then
    ret=0
  else
    ret=1
  fi
#  echo "Devel $ret, $N1 $N2"
  if [ $ret = 0 ]
  then
    ((AUTO_INDEX++))
  fi
#  echo "dev: $AUTO_INDEX $1"
  return $ret
}

function STATIC()
{
  cd $ROOT_BIN
  cp -a --parent .$1 $ROOT_STA
  rm -rf .$1
}

function AUTODEVEL()
{
  if [ -d $ROOT_BIN/$INCLUDEDIR ]
  then
    DEVEL "$INCLUDEDIR"
  fi

  # normal
  if [ "x$1" == "x" ]; then
    for i in `find $ROOT_BIN$LIBDIR -name '*.a'`
    do
      k=${i%.a}
      if [ -h "$k.so" -o -f "$k.so" ]
      then
        l=${i#$ROOT_BIN}
        STATIC "$l"
        AUTO_INDEX=-100
      else
        l=${i#$ROOT_BIN}
        DEVEL "$l"
        AUTO_INDEX=-100
      fi
    done
  fi

  # static_remove
  if [ "x$1" == "xstatic_remove" ]; then
    find $ROOT_BIN/$LIBDIR/*/ -name *.a -exec rm -f {} \;

    ls $ROOT_BIN/$LIBDIR/*.a &>/dev/null
    [ $? -eq 0 ] && STATIC "$LIBDIR/*.a"

  fi

  ###
  if [ -d $ROOT_BIN/$LIBDIR/pkgconfig ]
  then
    DEVEL "$LIBDIR/pkgconfig"
  fi
  if [ -d $ROOT_BIN/$DATADIR/aclocal ]
  then
    DEVEL "$DATADIR/aclocal"
  fi
  ls $ROOT_BIN/$BINDIR/$PROGRAM*-config &>/dev/null
  if [ $? -eq 0 ]
  then
    DEVEL "$BINDIR/$PROGRAM*-config"
  fi
  if [ -d $ROOT_BIN/$DATADIR/gtk-doc ] 
  then
    DEVEL "$DATADIR/gtk-doc"
  fi
  if [ -d $ROOT_BIN/$MANDIR/man3 ] 
  then
    DEVEL "$MANDIR/man3"
  fi
  if [ -d $ROOT_BIN/$MANDIR/man2 ]
  then
    DEVEL "$MANDIR/man2"
  fi
  if [ -d $ROOT_BIN/$LIBDIR/*/include ] 
  then
    DEVEL "$LIBDIR/*/include"
  fi
  if [ -d $ROOT_BIN/$LIBDIR/*/proc ]
  then
    DEVEL "$LIBDIR/*/proc"
  fi
  if [ -d $ROOT_BIN/$DATADIR/devhelp ] 
  then
    DEVEL "$DATADIR/devhelp"
  fi
#  echo "AUTO_INDEX: $AUTO_INDEX"
  if [ $AUTO_INDEX == 1 ]
  then
    # back all from autodevel to root_bin
    cd $ROOT_DEV
    cp -a --parent . $ROOT_BIN
    cd - > /dev/null
    rm -rf $ROOT_DEV/*
  fi
  return 0;
}

};
  print OUT q{
function install_source()
{
UNPACKED="yes"
dir=${dest%$(basename "$dest")}
if [ -n "$dir" ];
then
  echo "* Making directory (dest=$dest): $dir"
  mkdir -p "$dir"
fi
mkdir -p source_temp
cd source_temp
CODE="0"
if [ "$AUTO" == "yes" ]
then
case "$FILE" in
  *.tar)
    tar xvfO $STARTDIR/$PACKAGESRC "sources/$FILE" | tar xfv -
    CODE=$?
    UNPACKED="yes"
  ;;
  *.tar.gz)
    tar xvfO $STARTDIR/$PACKAGESRC "sources/$FILE" | tar xfvz -
    CODE=$?
    UNPACKED="yes"
  ;;
  *.tgz)
    tar xvfO $STARTDIR/$PACKAGESRC "sources/$FILE" | tar xfvz -
    CODE=$?
    UNPACKED="yes"
  ;;
  *.tar.bz2)
    tar xvfO $STARTDIR/$PACKAGESRC "sources/$FILE" | tar xfvj -
    CODE=$?
    UNPACKED="yes"
  ;;
  *.tbz)
    tar xvfO $STARTDIR/$PACKAGESRC "sources/$FILE" | tar xfvj -
    CODE=$?
    UNPACKED="yes"
  ;;
  *.zip)
    cd $SRCROOT
    tar xvf $STARTDIR/$PACKAGESRC "sources/$FILE"
    CODE=$?
    cd -
    if [ "x$CODE" == "x0" ]
    then
      unzip $SRCROOT/sources/$FILE
      CODE=$?
    fi
    UNPACKED="yes"
  ;;
  *)
    tar xvf $STARTDIR/$PACKAGESRC "sources/$FILE"
    cd ..
    mv "source_temp/sources/$FILE" $dest
    rmdir source_temp/sources 2>/dev/null
    cd -
    UNPACKED="no"
  ;;
esac
else
    tar xvf $STARTDIR/$PACKAGESRC "sources/$FILE"
    cd ..
    mv "source_temp/sources/$FILE" $dest
    cd -
    UNPACKED="no"
fi
cd ..

if [ "x$CODE" != "x0" ]
then
  echo "* Error in source $FILE or in package!"
  echo "* Fatal error!"
  CLEAN
  exit $CODE;
fi

if [ "x$UNPACKED" == "xyes" ]
then
  j=0
  file=""
  if [ "x$root" == "x" ]
  then
      cd source_temp
      files=`find * -mindepth 0 -maxdepth 0`
      for i in $files
      do
        let "j++"
	file="$i"
      done
      cd ..
  fi
  [ "$j" == "1" ] && root="$file"
  echo "* USING ROOT=$root"
  oldroot="$root"
  [ "x$root" == "x." ] && root=""
  if [ -d "source_temp/$root" ]
  then
    echo "PWD = $PWD"
    # it's needed because of bug in find command (moving glibc-2.3.2 - conform)
    if [ "x$dest" != "x" ]
    then
      echo "* Creating dir \"$dest\""
      mkdir -p "$dest"
    fi
    echo "* Moving... source_temp/$root/* to $dest"
    find source_temp/$root/ -maxdepth 1 -mindepth 1 -exec mv -f {} "$dest" \;
    rmdir "source_temp/$root/"
  else
    echo "* Directory source_temp/$root does not exists!"
    echo "* Propably root ($root) is wrong!"
    echo "* Fatal error!"
    CLEAN
    exit 1
  fi
  root="$oldroot"
fi
}

};

  print OUT q{
mkdir -p $TMPDIR/builddir
mkdir -p $TMPDIR/sandbox
mkdir -p $ROOT_BIN $ROOT_DEV $ROOT_STA
mkdir -p $SRCROOT
cd $TMPDIR/sandbox

cat > depmod << EOF
#!/bin/sh
echo "SANDBOX: depmod"
touch $TMPDIR/builddir/depmod.txt
exit 0
EOF
chmod 755 depmod

cat > scrollkeeper-update << EOF
#!/bin/sh
echo "SANDBOX: scrollkeeper-update"
touch $TMPDIR/builddir/scrollkeeper-update.txt
exit 0
EOF
chmod 755 scrollkeeper-update

cat > uname << EOF
#!/bin/sh
if [ "x\$1" = "x-r" ]
then
    if [ -n \$LINUX ]
    then
	/bin/uname \$@
    else
	echo $LINUX
    fi
else
    /bin/uname \$@
fi
EOF
chmod 755 uname


cat > update-mime-database << EOF
#!/bin/sh
echo "SANDBOX: update-mime-database"
touch $TMPDIR/builddir/update-mime.txt
exit 0
EOF
chmod 755 update-mime-database

cat > update-desktop-database << EOF
#!/bin/sh
echo "SANDBOX: update-desktop-database"
touch $TMPDIR/builddir/update-mime.txt
exit 0
EOF
chmod 755 update-desktop-database

cat > ldconfig << EOF
#!/bin/sh
echo "SANDBOX: ldconfig"
touch $TMPDIR/builddir/ldconfig.txt
exit 0
EOF
chmod 755 ldconfig

cat > gtk-update-icon-cache << EOF
#!/bin/sh
echo "SANDBOX: gtk-update-icon-cache"
exit 0
EOF
chmod 755 gtk-update-icon-cache

# pkg-config requires autodetection
cat > pkg-config << EOF
#!/usr/bin/perl

\$was = 0;
if ((\$ARGV[0] eq "--exists") || (\$ARGV[0] =~ "--atleast-version"))
{
  \$was = 1;
  open(TMP,">>$TMPDIR/builddir/pkg-config.txt");
  if (\$ARGV[0] eq "--exists") 
  {
  for (\$i=1; \$i<= \$#ARGV; \$i++)
  {
    \$el = \$ARGV[\$i];
    \$el =~ s/\t/\*/g;
    \$el =~ s/\s/\*/g;
    \$el =~ s/\,/\*/g;
    \$el =~ s/[\*]+/ /g;
    \$el =~ s/^\s//;
    if (\$el !~ /^\-\-/) {
      print TMP "\$el ";
    }
  }
  } else {
      if (\$ARGV[0] =~ /--atleast-version=(.+)/) {
        print TMP "\$ARGV[1] >= \$1 ";
      }
  }
  print TMP "|";
  close(TMP);
}
\$cmd = \$ARGV[0];
\$args = \$ARGV[1];
\$cmdline = '';
foreach (@ARGV) {
  \$cmdline = "\$cmdline \'" . \$_ . "\' ";
}

\$code = (0xffff & system "/usr/bin/pkg-config \$cmdline") >> 8;
if (\$was == 1)
{
    open(TMP,">>$TMPDIR/builddir/pkg-config.txt");
    print TMP "\$code\n";
    close(TMP);
}
exit \$code
EOF
chmod 755 pkg-config


cat > pkg-config-anal << EOF
#!/usr/bin/perl

\$el0 = 0;
\$el1 = 0;
open(TMP,"$TMPDIR/builddir/pkg-config.txt");
while (<TMP>)
{
    chomp;
    @col = split(/\|/);
    if (\$col[1] eq '0')
    {
      \$tab0[\$el0][0]=\$col[0];
      \$tab01[\$el0][1]=\$col[1];
      \$el0++;
    }
    if (\$col[1] eq '1')
    {
      \$tab1[\$el1][0]=\$col[0];
      \$tab1[\$el1][1]=\$col[1];
      \$el1++;
    }
}
close(TMP);

my @requires = ();
for (\$i=0; \$i<\$el0; \$i++)
{
  my \$line = \$tab0[\$i][0];
  my @tab = split(/[\s]+/,\$line);
  my \$require = '';
  foreach (my \$ind = 0; \$ind <= \$#tab; \$ind++) {
    my \$el_next = '';
    my \$el = \$tab[\$ind];
    \$require .= "\$el ";

    if (\$ind < \$#tab) {
      \$el_next = \$tab[\$ind+1];
    }
    if ((\$el ne ">" && \$el ne "<" && \$el ne "="  && \$el ne ">=" && \$el ne "<=") &&
        (\$el_next ne ">" && \$el_next ne "<" && \$el_next ne "="  && \$el_next ne ">=" && \$el_next ne "<=")) {
	\$require =~ s/(.+)\s\$/\$1/;
	push @requires, \$require;
	\$require = '';
    }
  }
}

my %req;
foreach (sort @requires) {
  # after sorting we get only "last" item as a "better"
  my @line = split(/[\s]/);
  \$req{\$line[0]} = \$line[0];
  if (@line > 1) {
    \$req{\$line[0]} .= "\t". \$line[2];
  }
}

foreach my \$key (sort keys %req) {
  print "pkgconfig." . \$req{\$key} . "\n";
}

#  print "<!--\n" if \$el1 != 0;
for (\$i=0; \$i<\$el1; \$i++)
{
  my \$commented = \$tab1[\$i][0];
  \$commented =~ s/\s>=\s/\t/;
  print "#pkgconfig." . \$commented ."\n";
}
#  print "-->\n" if \$el1 != 0;
EOF
chmod 755 pkg-config-anal

# fake chgrp to avoid error on non priviliged user
cat > chgrp << EOF
#!/usr/bin/perl

print "SANDBOX: chgrp\n";

if (\$ENV{USER} eq 'root')
{
  \$cmd = "/bin/chgrp ";
  foreach (@ARGV) {
    \$cmd .= "\$_ ";
  }
  system "\$cmd";
}

\$root_bin = "$ROOT_BIN";
\$i = 0;
\$mode = '';
foreach \$arg (@ARGV)
{
  chomp(\$arg);
  if (\$i == 0) {
    \$mode = \$arg;
  } else {
    if (\$arg =~ /(\$root_bin\/)(.+)/) {
      \$file = \$2;
    } else {
      \$file = \$arg;
    }
    push @files, \$file;
  }
  \$i++
}

if ((\$mode ne 'root') && (\$mode ne '0')) {
  open(PERMS,">>$TMPDIR/builddir/perms.txt");
  foreach (@files)
  {
    print PERMS "root:\$mode\t\$_\n";
  }
  close(PERMS);
}

exit 0

EOF
chmod 755 chgrp

# fake chown to avoid error on non priviliged user
cat > chown << EOF
#!/usr/bin/perl

print "SANDBOX: chown\n";

if (\$ENV{USER} eq 'root')
{
  \$cmd = "/bin/chown ";
  foreach (@ARGV) {
    \$cmd .= "\$_ ";
  }
  system "\$cmd";
}

\$root_bin = "$ROOT_BIN";
\$i = 0;
\$mode = '';
foreach \$arg (@ARGV)
{
  chomp(\$arg);
  if (\$i == 0) {
    \$mode = \$arg;
  } else {
    if (\$arg =~ /(\$root_bin\/)(.+)/) {
      \$file = \$2;
    } else {
      \$file = \$arg;
    }
    push @files, \$file;
  }
  \$i++
}

if (\$mode =~ /(.*)[\.:](.*)/)
{
  \$user = \$1;
  \$group = \$2;
} else
{
  \$user = \$mode;
}

if (!(((\$user eq 'root') && (\$group eq '')) || 
    ((\$group eq 'root') && (\$user eq '')) ||
    ((\$group eq 'root') && (\$user eq 'root'))))
{
open(PERMS,">>$TMPDIR/builddir/perms.txt");
foreach (@files)
{
  if (\$user eq '') {
    \$user = 'root';
  }
  if (\$group eq '') {
    \$group = 'root';
  }
  print PERMS "\$user:\$group\t\$_\n";
}
close(PERMS);
}
exit 0
EOF
chmod 755 chown

# fake install to avoid e.g. 'install -o root -g root'
cat > install << EOF
#!/usr/bin/perl

\$root_bin = "$ROOT_BIN";
\$last = '';
\$cmd = "/bin/install ";
\$directory = 0;
foreach \$arg (@ARGV)
{
  chomp(\$arg);
  \$print = 1;
  if (\$last eq '-o' || \$last eq '-g') {
    \$print = 0;
  }
  if (\$arg eq '-d')
  {
    \$directory = 1;
  }
  if (\$print == 1 && \$arg ne '-o' && \$arg ne '-g')
  {
      \$cmd .= "'\$arg' ";
  }
  if (\$print == 0 && \$last eq '-o')
  {
    \$user = \$arg;
  }
  if (\$print == 0 && \$last eq '-g')
  {
    \$group = \$arg;
  }
  if ((\$arg !~ /^\-/) && ((\$last !~ /^\-[^s]/) || ((\$last =~ /^\-/) && (\$directory == 1) )) && (\$last ne '') && (\$user ne '') && (\$group ne ''))
  {
    if (\$last =~ /^\-/) {
      if (\$last =~ /^\-s/) {
        push @buffer, "\$arg";
      } else {
        push @files, "\$arg";
      }
    } #else {
      if (\$arg !~ /^\$root_bin/)
      {
        my @array = split(/\//, \$arg);
	my \$el = \$array[\$#array];
        push @buffer, \$el;
      } else {
        foreach \$buf (@buffer) {
	  my \$file = "\$arg/\$buf";
	  if (\$file ne \$files[\$#files]) {
            push @files, \$file;
	  }
	}
        @buffer = ();
      }
    #}
  }
  \$last = \$arg;
}


if (!(((\$user eq 'root') && (\$group eq '')) || 
    ((\$group eq 'root') && (\$user eq '')) ||
    ((\$user eq 'root') && (\$group eq 'root')) ||
    ((\$user eq '') && (\$group eq ''))))
{
#foreach (@ARGV)
#{
#  chomp;
#  print PERMS "\$_ ";
#}
#print PERMS "\n";

open(PERMS,">>$TMPDIR/builddir/perms.txt");

foreach my \$file (@files)
{
  \$file =~ s/\$root_bin\///;
  my \$full_file = "\$root_bin/\$file";
  if (-f "\$full_file" || -d "\$full_file") {
    print PERMS "\$user:\$group\t\$file\n";
  }
}
close(PERMS);
}

if (\$ENV{USER} eq 'root')
{
  \$cmd2 = "/bin/install ";
  for (my \$i=0; \$i < \$#ARGV; \$i++) {
    \$cmd2 .= "'\$ARGV[\$i]' ";
  }
  if (\$ARGV[\$#ARGV] =~ /^\/etc/ || \$ARGV[\$#ARGV] =~ /^\/usr/ || \$ARGV[\$#ARGV] =~ /^\/lib/ ) {
    \$autoprefix = "\$root_bin";
  } else {
    \$autoprefix = "";
  }
  \$cmd2 .= "'\$autoprefix\$ARGV[\$#ARGV]'";
  
  print "SANDBOX: \$cmd2\n";
  system "\$cmd2";
} else {
  print "SANDBOX: \$cmd\n";
  system "\$cmd";
}

EOF
chmod 755 install

PATH=$PWD:$PATH
cd $TMPDIR/builddir

function_create_fhs "$ROOT_BIN"
function_create_fhs "$ROOT_DEV"
function_create_fhs "$ROOT_STA"
};

  my @sources = xpath("$FILE",'/box/source/@id');

  foreach my $src (@sources)
  {
    my @patches = xpath("$FILE","/box/source[\@id = \"$src\"]/patch");
    my $xpath = XML::XPath->new(filename => "$FILE");
    my $nodes = $xpath->findnodes("/box/source[\@id = \"$src\"]/patch");
    my @levels;
    my $l = 0;
    foreach my $node ($nodes->get_nodelist)
    {
      $levels[$l] = $xpath->findvalue("\@level", $node);
      if (length($levels[$l]) == 0) {
        # default level
        $levels[$l] = 1;
      }
      $l++;
    }

    my @urls = xpath("$FILE","/box/source[\@id = \"$src\"]/url");
    @temp = xpath("$FILE","/box/source[\@id = \"$src\"]/\@root");
    my $root = getfirst(@temp);
    $root = macro($root, $V, $P);
    my @dests = xpath("$FILE","/box/source[\@id = \"$src\"]/\@dest");
    my $dest = getfirst(@dests);
    $dest = macro($dest, $V, $P);
    my @autos = xpath("$FILE","/box/source[\@id = \"$src\"]/\@auto");
    my $auto = getfirst(@autos);
    print OUT "\n";
    print OUT "# Source: $src\n";
    print OUT "root=\"$root\"\n";
    print OUT "dest=\"$dest\"\n";
    print OUT "# auto: \'$auto\'\n";
    print OUT "AUTO=yes\n";
    if ($auto eq "no")
    {
      print OUT "AUTO=no";
    }
    print OUT q{
[ "x$dest" == "x" ] && dest="."
};
    foreach (@urls)
    {
      my $base = macro(basename($_), $V, $P);
      my $name = prefixname($base);
      print OUT "# source: $base\n";
      print OUT "SOURCE_$src=$base\n";
      print OUT "FILE=$base\n";
      print OUT "NAME=$name\n";
    }
    print OUT q{
#[ "x$root" == "x" ] && root="$NAME"
install_source
};
    my $i = 0;
    my $autoreconf = 0;
    foreach my $patch (@patches)
    {
      my $level = $levels[$i];
      my $param = "-p$level";
      print OUT "# patch: $level, $patch\n";
      print OUT "cd \"\$dest\";\n";
      print OUT "echo \"* patching: $patch with level $level\"\n";
      print OUT "tar xfO \$STARTDIR/\$PACKAGESRC patches/$patch  | patch $param;\n";
      print OUT 'CODE=$?',"\n";
      print OUT "cd -\n";
      print OUT 'if [ "x$CODE" != "x0" ]; then echo "* Fatal error!"; CLEAN; exit $CODE;  fi',"\n";
      print OUT "PATCH_LIST=\"\$PATCH_LIST $patch\"\n";
      $i++;
    }
  }
    print OUT q{
PATCH_DETECTED="NO"
for i in $PATCH_LIST
do
    tar xfO $STARTDIR/$PACKAGESRC patches/$i | grep -E '\+\+\+.*(configure.in|configure.ac|Makefile.am|.m4)' 2>&1 >/dev/null
    if [ $? == 0 ]
    then
      PATCH_DETECTED="yes"
    fi
done
};
  print OUT q{
# run autoreconf if patches changed configure.{in,ac} or Makefile.am
if [ "x$NO_AUTOTOOLS" == "x" ]
then
  if [ "$PATCH_DETECTED" == "yes" ] 
  then
      echo "* Running autoreconf -vif"
      autoreconf -vif
  fi
fi
AUTO_INDEX=0
  };
  my @code = xpath("$FILE",'/box/build/code');
  foreach my $line (@code)
  {

    $line = macro($line, $V, $P);
#    print OUT "cat >> /dev/stdout << ENDOFFILE\n";
#    print OUT "* $line\n";
#    print OUT "ENDOFFILE\n";
    print OUT "$line\n";
  }
  print OUT q{
# run autodevel allways!
if [ "x$NO_AUTODEVEL" == "x" ]
then
  AUTODEVEL
fi
  };
  my @list;
  @list = xpath("$FILE","/box/package/provides/\@value");
  print OUT "provides='@list'\n";
  @list = xpath("$FILE","/box/package/requires/\@value");
  print OUT "requires='@list'\n";
  print OUT q{
  echo "package.$PROGRAM	$VERSION" >>  ../builddir/provides.txt
  for i in $provides
  do
    echo "$i" >>  ../builddir/provides.txt
  done
  for i in $requires
  do
    echo "$i" >>  ../builddir/requires.txt
  done
};
  ### End of Generation
  system "rm -f $FILE";

  print OUT q{
# EXPERIMENTAL - remove libtool .la files when we using pkgconfig .pc files
# this should stop stupid libtool linking poprgation eg. libhowl
#
# this should be before AUTODEVEL and DEVEL :(
# now assume that .pc files are in DEVEL package
# and libtool .la files are in normal
REMOVE_LA='no'
if [ -d $ROOT_DEV/usr/lib/pkgconfig ]; then
for i in $ROOT_DEV/usr/lib/pkgconfig/*.pc
do
  REMOVE_LA='yes'
#  i=${i#usr/lib/pkgconfig/}
#  i=${i%.pc}
#  la1="usr/lib/$i.la"
#  la2="usr/lib/lib$i.la"
#  if [ -f "$la1" ]; then
#    echo "Removing: $la1"
#  fi
#  if [ -f "$la2" ]; then
#    echo "Removing: $la2"
#  fi
done
fi

if [ $REMOVE_LA == "yes" ]
then
  for i in $ROOT_BIN/usr/lib/*.la
  do
    if [ -f "$i" ]
    then
      name=${i#$ROOT_BIN/}
      echo "Removing: $name"
      rm -f "$i"
    fi
  done
fi

function write_build_provides
{
DIR="$1";
if [ -d "$DIR$LIBDIR/pkgconfig" ]
then
  for i in `find $DIR$LIBDIR/pkgconfig -name '*.pc'`
  do
    j=${i#$DIR$LIBDIR/pkgconfig/}
    k=${j%.pc}
    v=$(grep "Version: " $i | cut -d ' ' -f2)
    echo -e "pkgconfig.$k\t$v" >> ../builddir/buildprovides.txt
  done
fi
}

function write_provides
{
DIR="$1";
F1=`find $DIR$LIBDIR -name '*.dll' -maxdepth 1 2>/dev/null`
F2=`find $DIR$LIBDIR/mono/gac -name '*.dll' 2>/dev/null`
F="$F1 $F2"
for i in $F
do
  n=$(mono-find-provides "$i" | cut -d '(' -f2 | cut -d ')' -f1)
  v=$(mono-find-provides "$i" | cut -d ' ' -f3)
  echo -e "dotnet.$n.dll\t$v" >> ../builddir/provides.txt
done

cat >> findprovides.pl << EOF;
#!/usr/bin/perl

  my \$dir = "$DIR";
  my %sonames;
  my %fsonames;
  my @changes1 = ();
  my @changes2 = ();
  my @sonamekeys = ();
  my @fsonamekeys = ();
  
  open(F, "find \$dir |");
  my @files = <F>;
  close(F);
  
  foreach my \$path (@files) {
    chomp(\$path);
    if (\$path =~ /(usr\/)?lib\/(.*\.so.*)\$/) {
	if (-f "\$path") {
	    my \$wassomething = 0;
	    my \$libname = \$2;
	    my \$depth = 0;
	    open(CMD,"readelf -d \$path 2>/dev/null|");
	    my \$s = '';
	    while (<CMD>) {
	      if (/soname.+\[(.+)\]/) {
	        \$s = \$1;
		\$wassomething = 1;
	      }
	    }
	    if (\$libname =~ /\/.*\//) {
		\$depth = 1;
	    }
	    
            if (\$wassomething == 1 &&
	        (index(\$libname, '/') < 0 ||
	        (index(\$libname, '/') == rindex(\$libname, '/')) ||
		\$depth == 1)) 
	    {
		if (\$s =~ /^lib/) {
	    	    \$sonames{\$s} = 1;
		}
	    }
									
	    close(CMD);
	    # sometimes library doesn't have "soname" in readelf -d
	    if (\$wassomething == 0) {
	      if (\$path =~ /(usr\/)?lib\/(lib[^\/]*\.so)\$/) {
	        my \$lib = \$2;
	        open(FILE,"file -L \$path 2>/dev/null|");
		my \$line = <FILE>;
		close(FILE);
		if (\$line =~ /shared object/) {
	          \$sonames{\$lib} = 1;
		}
	      }
	    }
	}
    }
  }
  if (keys %sonames) {
    open(LOG, ">>../builddir/provides.txt");
    foreach (keys %sonames) {
      print LOG "soname.\$_\n";
    }
  }
  close(LOG);
  system "rm -f findprovides.pl";
EOF
  perl findprovides.pl
}

# FIXME: add soname requires, add mono requires
function write_requires
{
DIR="$1";
cat >> findrequires.pl << EOF;
#!/usr/bin/perl

sub dirname
{
  my @array = split(/\//, \$_[0]);
  my \$result = '';
  for (my \$i = 0; \$i < \$#array; \$i++) {
    \$result .= \$array[\$i] . "/";
  }
  return \$result;
}

  my \$dir = "$DIR";
  my %sonames2;
  my %fsonames2;
  my %nonstandard;

  open(F, "find \$dir |");
  my @files = <F>;
  close(F);

  # detect libraries in non standard directories
  foreach my \$path (@files) {
    chomp(\$path);
    if ((\$path =~ /(usr\/)?lib\/.*so.*\$/) || (\$path =~ /(usr\/)?s?bin\/.+\$/) ||
        (\$path =~ /usr\/libexec\/.+\$/)) {
	    # if non standard directory and required library is in the
	    # same directory then don't add this as "required"
	    my \$lib = \$path;
	    if (\$lib !~ /(s?bin|usr\/s?bin|lib|usr\/lib|usr\/libexec)\/[^\/]+\$/) {
	      # non standard dir
	      my \$basename = dirname(\$lib);
	      \$nonstandard{\$basename} = 1;
	      #print "Non standard: \$basename\n";
	    }
        }
  }
  foreach my \$path (@files) {
    chomp(\$path);
    # mono .NET stuff
    if (\$path =~ /\.(exe|dll)\$/) {
      open(CMD,"mono-find-requires \$path 2>/dev/null|");
      while (<CMD>) {
        if (/mono\((.+)\)\s\=\s(.+)\$/) {
	  my \$fullname = "\$1.dll \$2";
	  # if library in the same dir then don't add this
	  my \$file = dirname(\$path) . "\$1.dll";
	  my \$washere = 0;
	  foreach my \$path2 (@files) {
	    chomp(\$path2);
	    if (\$file =~ /\Q\$path2\E\$/) {
	      \$washere = 1;
	    }
	  }
	  if (\$washere == 0) {
	    \$sonames2{"dotnet.\$fullname"} = 1;
	  }
        }
      }
      close(CMD);
    }
    # soname stuff
    if ((\$path =~ /(usr\/)?lib\/.*so.*\$/) || (\$path =~ /(usr\/)?s?bin\/.+\$/) ||
        (\$path =~ /usr\/libexec\/.+\$/)) {
	if (-f "\$path") {
	    open(CMD,"readelf -d \$path 2>/dev/null|");
	    while (<CMD>) {
	      if (/Shared\slibrary.+\[(.+)\]/) {
	        my \$libname = \$1;
		my \$pathdirname = dirname(\$path);

	  my \$file = dirname(\$path) . "\$1";
	  my \$washere = 0;
	  foreach my \$path2 (@files) {
	    chomp(\$path2);
	    if (\$file =~ /\Q\$path2\E\$/) {
	      \$washere = 1;
	    }
	  }
	  if (\$washere == 0) {
	    \$sonames2{"soname.\$libname"} = 1;
	  }
	      }
	    }
	    close(CMD);
	}
    }
  }
  open(LOG, ">>../builddir/requires.txt");
  foreach (keys %sonames2) {
    s/\s/\t/;
    print LOG "\$_\n";
  }
  close(LOG);
  system "rm -f findrequires.pl";
EOF
  perl findrequires.pl
}

write_build_provides "$ROOT_BIN"
write_build_provides "$ROOT_DEV"

# write_build_requires <- in other place implemented!

write_provides "$ROOT_BIN"
write_provides "$ROOT_DEV"

write_requires "$ROOT_BIN"
write_requires "$ROOT_DEV"

function_clean_langs_default
function_fix_symlinks "$ROOT_BIN"
function_fix_symlinks "$ROOT_DEV"

function_clean_fhs "$ROOT_BIN"
function_clean_fhs "$ROOT_DEV"
function_clean_fhs "$ROOT_STA"

if [ -f "../builddir/pkgconfig-fail.txt" ]
then
  echo "Missing requires:"
  cat "../builddir/pkgconfig-fail.txt" | sort -u > "../builddir/pkgconfig-fail2.txt"
  while read n v
  do
    echo -ne "$n\t$v\t\t"
    box -what_provide pkgconfig.$n $v
  done < ../builddir/pkgconfig-fail2.txt
fi


NPACKAGES="0"
MAKEBOX $ROOT_BIN "" "i686"
NPACKAGES=$(($NPACKAGES+$IS_PACKAGE))
MAKEBOX $ROOT_DEV "devel" "i686"
NPACKAGES=$(($NPACKAGES+$IS_PACKAGE))
MAKEBOX $ROOT_STA "static" "i686"
NPACKAGES=$(($NPACKAGES+$IS_PACKAGE))
echo "------------------------------------------------------------------------------"
#echo "PACKAGES: $NPACKAGES"
#CLEANCLEAN
CLEAN
if [ "$NPACKAGES" == "0" ]
then
  echo "* ERROR: there is no package!"
  exit 1
fi
};
  close(OUT);
  my $rc;
  my $alltime, my $t1, my $t2;
  $t1 = time;
  my @starttime = localtime($t1);
  # touch source package!
  utime $t1, $t1, $file;
  my $timestr = format_time(@starttime);
  my $logdir = "$LOGDIR/trunk/$timestr" . "-$P";
  my $logfile = "$logdir/$P-$V-$R.log";
  my $shortlogfile = "$timestr" . "-$P/$P-$V-$R.log";
  system "mkdir -p $logdir";
  print "Building $P...\n";
  $rc = 0xffff & system "cd $LOGDIR/trunk/; ln -sf $shortlogfile last.log; cd - >/dev/null; sh $P.sh '$timestr' 2>&1 | tee /dev/tty12 &> $logfile";
  $rc = $rc >> 8;
  $t2 = time;
  $alltime = sprintf("%d",1+($t2-$t1)/60);
  my $sec = sprintf("%d",($t2-$t1) % 60);
  open(LOG,">>$logfile");
  print LOG "* BUILDTIME: $alltime" ."m$sec" . "s\n";
  print LOG "* EXIT CODE: $rc\n";
  close(LOG);
  system "gzip $logfile";
  system "cd $LOGDIR/trunk/; ln -sf $shortlogfile.gz last.log";
  exit $rc;
}

# ============================================================================
# list_package
# ============================================================================

sub list_package
{
  my $file = $_[0];
  my $type = "";
  my @array = split(/\./, $file);
  my $last = $#array;
  my @files = ();
  if ($last < 2) {
    $type = "";
  } else {
    $type = $array[$#array - 1];
  }
  if (-f $file)
  {
    if (($array[$last] eq "box") and ($type ne "src")) {
      my $code = system "tar xfO $file package.box &>/dev/null";
      if ($code == 0)
      {
        $file =~ /(.*)\.box$/;
	my $datafile = "$1.tar.gz";
	my @arr = split(/\//, $datafile);
	$datafile = $arr[$#arr];
	open(CACHE,"tar xfO $file cache/files 2>/dev/null|");
	@files = <CACHE>;
	close(CACHE);
	if (scalar(@files) == 0) {
          open(FILES,"tar xfO $file data/$datafile | tar tfz -|");
	  foreach (<FILES>) {
	    chomp;
	    if (/[^\/]$/) {
	      push @files, $_;
	    }
	  }
	  close(FILES);
	}
	return sort @files;
      } else
      {
        print "File $file is not binary package!\n";
	return ();
      }
    } else
    {
      print "File $file is not binary package!\n";
      return ();
    }
  } else
  {
    print "File $file does not exists!\n";
    return ();
  }
}

# ============================================================================
# install_package(file, root)
# ============================================================================

sub install_package
{
  my $file = $_[0];
  my $root = $_[1];
  my @temp = ();
  my $V;
  my $P;
  my $rc;

  if (! -f $file ) { die "File $file not found!\n"; }

  $rc = 0xffff & system "tar xf $file package.box &>/dev/null";
  if ($rc !=0) { die "Error: cannot get package.box from package!\n"; }
  system "tar xf $file cache/post-install &>/dev/null";  
  @temp = xpath("package.box",'/box/source/name/@version');
  $V = getfirst(@temp);
  @temp = xpath("package.box",'/box/source/name');
  $P = getfirst(@temp);
  my @PREINSTALL = xpath("package.box",'/box/package/pre-install');
  if (@PREINSTALL) {
    @PREINSTALL = split("\n", $PREINSTALL[0]);
  }
  my @POSTINSTALL = xpath("package.box",'/box/package/post-install');
  if (@POSTINSTALL) {
    @POSTINSTALL = split("\n", $POSTINSTALL[0]);
  }
  my @FI;
  if (-f "cache/post-install")
  {
    open(F, "cache/post-install");
    while (<F>)
    {
      chomp;
      push @POSTINSTALL, $_;
    }
    close(F);
  }
  
  system "rm -f package.box";
  system "rm -f cache/post-install; rmdir cache &>/dev/null";

  open(OUT,">$P-install.sh");
  ### Generatin script for install package ###
  print OUT "#!/bin/bash\n";
  print OUT "# This file is generated please do not edit!\n";
  print OUT '# $Id: box.pl,v 1.167 2006/01/23 19:38:47 kaliber Exp $',"\n";
  print OUT "# file: $file\n\n";
  print OUT "LC_ALL=C; LANG=C;\n";
  foreach (@PREINSTALL)
  {
    print OUT "$_\n";
  }
  my @schemas = ();
  open(SCH,"tar xfO $file data/$P*-$V-*.tar.gz | tar fzt -|");
  while (<SCH>)
  {
    if (/.*etc\/gconf\/schemas\/(.+).schemas/)
    {
      push @schemas, $1;
    }
  }
  close(SCH);
  my $n_schemas = scalar(@schemas);
  print OUT "ROOT=$root\n";
  print OUT "for i in `tar xfO $file data/$P*-$V-*.tar.gz | tar  tvfz - | grep '^-' | tr -s ' ' | cut -d ' ' -f6`\n";
  print OUT "do\n";
  print OUT "  rm -f \$ROOT/\$i\n";
  print OUT "done\n";
  print OUT "tar xfO $file data/$P*-$V-*.tar.gz | tar  xvfz - -C $root --overwrite \n";
  print OUT "tar xfO $file cache/perms 2>/dev/null > $P-perms.txt \n\n";
  print OUT "while read a b\n";
  print OUT "do\n";
  print OUT "if [ -f \"\$ROOT\$b\" -o -d \"\$ROOT\$b\" ]\n";
  print OUT "then\n";
  print OUT "  perms=\$(find \$ROOT\$b -printf '\%m' -maxdepth 0 2>/dev/null)\n";
  print OUT "  echo \"chown \$a  \$ROOT\$b\"\n";
  print OUT "  chown \$a  \$ROOT\$b 2>/dev/null\n";
  print OUT "  echo \"chmod \$perms \$ROOT\$b\"\n";
  print OUT "  chmod \$perms \$ROOT\$b 2>/dev/null\n";
  print OUT "fi\n";
  print OUT "done < $P-perms.txt\n";

  foreach my $line (@POSTINSTALL)
  {
    $line = macro($line, $V, $P);
    if ($line =~ /\%\{gconf_schemas_install\}/)
    {
      if ($n_schemas > 0)
      {
        print OUT 'export GCONF_CONFIG_SOURCE=`gconftool-2 --get-default-source`', "\n";
        foreach (@schemas)
	{
	  print OUT "echo \"* gconftool-2 --makefile-install-rule /etc/gconf/schemas/$_.schemas\"\n";
          print OUT "gconftool-2 --makefile-install-rule /etc/gconf/schemas/$_.schemas\n";
	}
      }
    } else
    {
      #print OUT "echo \'* $line\'\n";
      print OUT "$line\n";
    }
  }
  print OUT "rm -f $P-perms.txt\n";
  print OUT "rm -f $P-install.sh\n";
  close(OUT);
  $rc = 0xffff & system "sh $P-install.sh";
  $rc = $rc >> 8;
  print "EXIT: $rc\n";
}

# ============================================================================
# generate_todo_list(tag?)
# ============================================================================

sub generate_todo_list
{
  my $tag = $_[0];
  my $allfiles;
  my $V;
  my $P;
  my @files;
  my @temp;

  print "TODO:\n";
  if ($tag ne '0') {
    system "mkdir -p $TEMPDIR; cd $TEMPDIR; cvs co -r $tag $CVSBOXES";
    $allfiles = "$TEMPDIR/$CVSBOXES/*.box";
  } else
  {
    $allfiles = "$BOXES/*.box";
  }
  @files = glob($allfiles);
  foreach my $file (@files)
  {
    @temp = xpath("$file",'/box/source/name/@version');
    $V = getfirst(@temp);
    @temp = xpath("$file",'/box/source/name');
    $P = getfirst(@temp);
    @temp = xpath("$file",'/box/todo/do');
    if (@temp > 0 )
    {
      print "* $P\n";
    }
    foreach my $do (@temp)
    {
	print "- $do\n";
    }
    @temp = ();
  }
  if ($tag ne '0') {
    system "rm -rf $TEMPDIR/$CVSBOXES";
  }

}

# ============================================================================
# diff_package (file1, file2, OUT)
# ============================================================================

sub diff_package
{
  my $OUT;
  my $file;
  my $type = "";
  my @array;
  my $last;
  my @files1;
  my @files2;

  my @temp = ();
  my $V1;
  my $V2;
  my $diffignore1;
  my $diffignore2;
  my $diffignore1a;
  my $diffignore2a;

  if (!defined($_[2]))
  {
    $OUT = \*STDOUT;
  } else {
    $OUT = $_[2];
  }

  # first
  $file = $_[0];
  @array =  split(/\./, $file);
  $last = $#array;
  if ($last < 2) {
    $type = "";
  } else {
    $type = $array[$#array - 1];
  }
  if (-f $file)
  {
    if ($array[$last] eq "box" and $type ne "src") {
      my $code = system "tar xf $file package.box &>/dev/null";
      if ($code == 0)
      {
        @temp = xpath("package.box",'/box/source/name/@version');
        $V1 = getfirst(@temp);
        @temp = xpath("package.box",'/box/package/diffignore');
        $diffignore1 = getfirst(@temp);
        $diffignore1a = macro($diffignore1, $V1, '');
	system "rm -f package.box";

        $file =~ /(.*)\.box$/;
	my $datafile = "$1.tar.gz";
	my @arr = split(/\//, $datafile);
	$datafile = $arr[$#arr];
        open(LST, "tar xfO $file data/$datafile | tar tfz - | sort |");
	@files1 = <LST>;
	close(LST);
      } else
      {
        print $OUT "File $file is not binary package!\n";
      }
    } else
    {
      print $OUT "File $file is not binary package!\n";
    }
  } else
  {
    print $OUT "File $file does not exists!\n";
  }

  # second
  $file = $_[1];
  @array =  split(/\./, $file);
  $last = $#array;
  if ($last < 2) {
    $type = "";
  } else {
    $type = $array[$#array - 1];
  }
  if (-f $file)
  {
    if ($array[$last] eq "box" and $type ne "src") {
      my $code = system "tar xf $file package.box &>/dev/null";
      if ($code == 0)
      {
        @temp = xpath("package.box",'/box/source/name/@version');
        $V2 = getfirst(@temp);
        @temp = xpath("package.box",'/box/package/diffignore');
        $diffignore2 = getfirst(@temp);
        $diffignore2a = macro($diffignore2, $V2, '');
	system "rm -f package.box";

        $file =~ /(.*)\.box$/;
	my $datafile = "$1.tar.gz";
	my @arr = split(/\//, $datafile);
	$datafile = $arr[$#arr];
        open(LST, "tar xfO $file data/$datafile | tar tfz - | sort |");
	@files2 = <LST>;
	close(LST);
      } else
      {
        print $OUT "File $file is not binary package!\n";
      }
    } else
    {
      print $OUT "File $file is not binary package!\n";
    }
  } else
  {
    print $OUT "File $file does not exists!\n";
  }

  if ("$diffignore1" ne "$diffignore2")
  {
    print $OUT "Warning: different diffignore ($diffignore1, $diffignore2)\n";
  } else
  {
    if ("$diffignore1" ne '') {
      print $OUT "Warning: using diffignore ($diffignore1: '$diffignore1a'->'$diffignore2a')\n";
    }
  }
  open(TMP1, ">$TEMPDIR/boxa$$");
  foreach (@files1)
  {
    chomp;
    if (("$diffignore1" eq "$diffignore2") && (length($diffignore1) > 0))
    {
      s/$diffignore1a/$diffignore1/g;
    }
    print TMP1 "$_\n";
  }
  close(TMP1);

  open(TMP2, ">$TEMPDIR/boxb$$");
  foreach (@files2)
  {
    chomp;
    if (length($diffignore1) > 0) {
      s/$diffignore2a/$diffignore1/g;
    }
    print TMP2 "$_\n";
  }
  close(TMP2);

  open(DIFF,"diff -u $TEMPDIR/boxa$$ $TEMPDIR/boxb$$|");
  print $OUT <DIFF>;
  close(DIFF);
  system("rm -f $TEMPDIR/boxa$$ $TEMPDIR/boxb$$");
}

# ============================================================================
# validate_package (file1)
# ============================================================================
sub validate_package
{
  my $file;
  $file = $_[0];

  open(CMD,"xmllint --schema /usr/share/pgw-box/pgw-box.xsd $file 2>&1 | grep \"$file validates\"|");
  my @res = <CMD>;
  close(CMD);

  if (not @res) {
    return 1;
  } else  {
    return 0;
  }
}

# ============================================================================
# initialize()
# ============================================================================

sub makedir
{
  my ($directory) = (@_);
  print STDERR "Making $directory...\n";
  system "mkdir -p $directory\n";
}

sub initialize
{
  makedir($TEMPDIR);
  makedir($BOXES);
  makedir($DISTFILES);
  makedir($DISTFILES_EXTRA);
  makedir($PATCHES);
  makedir($LOGDIR);
  makedir($TESTSDIR);
  makedir("$BOXROOT/FACTORY/main");
  makedir("$BOXROOT/FACTORY/main_backup");
}

# ============================================================================
# template($url)
# ============================================================================

sub template
{
  my ($realurl) = (@_);
  my $name = 'template';
  my $version = '1.0';
  my $homepage = 'http://template.org/';
  my $url = 'http://template.org/template-%{version}.tar.gz';
  my $type = '';

  if (defined($realurl)) {
   $url = $realurl;
  }

  if ($realurl =~ /^http:\/\/([^\/]+)/)
  {
    $homepage = "http://$1/";
  }
  if ($realurl =~ /^ftp:\/\/ftp\.([^\/]+)/)
  {
    $homepage = "http://www.$1/";
  }
  my $filename = basename($realurl);
  if ($filename =~ /^(.+)-([0-9.a-zA-Z]+)\.(tar\.gz|tgz|tar\.bz2|zip)/)
  {
    $name = $1;
    $version = $2;
    my $ext = $3;
    my $dirname = dirname($realurl);
    # sourceforge.net
    if ($dirname =~ /http:\/\/.*dl\.(sf|sourceforge)\.net\/sourceforge\/([^\/]+)/)
    {
       $homepage = "http://$2.sourceforge.net/";
       $dirname =~ s/http:\/\/.*dl\.(sf|sourceforge)\.net\/sourceforge/\%\{SF\.NET\}/;
    }
    # sourceforge.net prdownloads
    if ($dirname =~ /http:\/\/.*prdownloads\.(sf|sourceforge)\.net\/([^\/]+)/)
    {
       $homepage = "http://$2.sourceforge.net/";
       $dirname =~ s/http:\/\/.*prdownloads\.(sf|sourceforge)\.net/\%\{SF\.NET\}/;
    }
    # gnome.org
    if ($dirname =~ /(f|ht)tp:\/\/.*gnome\.org\/.*sources\/([^\/]+)\/([^\/]+)/)
    {
       $homepage = "http://www.gnome.org/projects/$2/";
       $dirname =~ s/(f|ht)tp:\/\/.*gnome\.org\/.*sources\/([^\/]+)\/([^\/]+)/\%\{GNOME\.ORG\}\/$2\/\%\{main_version\}/;
    }
    # cpan.org
    if ($dirname =~ /(f|ht)tp:\/\/.*cpan\.org\//)
    {
       $type = 'CPAN.ORG';
       my $category = '';
       if ($filename =~ /([^\-]+)\-(.+)/) {
         $category = $1;
       }
       $homepage = "http://www.cpan.org/";
       $dirname =~ s/(f|ht)tp:\/\/.*cpan\.org\/.*\/([^\/]+)/\%\{CPAN\.ORG\}\/$category/;
    }
    
    $url = $dirname . $name . "-%{version}." . $ext;
    $name =~ tr/A-Z/a-z/;
    
    # special for cpan.org
    if ($type eq 'CPAN.ORG')
    {
       $name = "perl-$name";
    }
  }

  my $template = q{<?xml version="1.0" ?>
<box group="candidate">
<source id="main">
};
  $template .= "\t<name version=\"$version\" release=\"1\">$name</name>\n";
  $template .= "\t<url>$url</url>\n";
  $template .= "\t<homepage>$homepage</homepage>";
  $template .= q{
</source>

<build>
	<code><![CDATA[};
  if ($type eq 'CPAN.ORG') 
  {
    $template .= q{
perl Makefile.PL
%{make} &&
%{make_install}
};
  } else
  {
    $template .= q{
%{configure} &&
%{make} &&
%{make_install}
};
  }
  if ($name =~ /lib.+/)
  {
    $template .= q{};
  }
  $template .=  q{]]>
	</code>
</build>

<package name="normal">
	<files>
	</files>
</package>

<changelog><![CDATA[
$fuckincvs: $
]]>
</changelog>

</box>
};
 $template =~ s/fuckincvs/Log/m;
 print $template;
}


# ============================================================================
# check_factory(source_package)
# ============================================================================
sub check_factory
{
  my $file = $_[0];
  my $origfile = $file;
  my @temp = ();
  my $post_install;
  my $V;
  my $P;
  my $R;
  my $fV;
  my $fP;
  my $fR;
  my $TMPUNPACKED;
  my $TMPUNPACKEDOLD;

  if (! -f "$file") {
    die "File $file not found!\n";
  }
  my $FILE = "package.box";
  my $FFILE = "package.box";
  open(CMD,"tar xf$COMPRESS"." $file $FILE 2>/dev/null|");
  close(CMD);
  if (! -f $FILE) {
    die "Bad file!\n";
  }
  system "mv $FILE package_new.box";
  $FILE = "package_new.box";
  if (validate_package($FILE) == 1) {
    #print STDERR "\n* XML file is not valid box file!\n";
  }

  @temp = xpath("$FILE",'/box/source/name/@version');
  $V = getfirst(@temp);
  @temp = xpath("$FILE",'/box/source/name/@release');
  $R = getfirst(@temp);
  @temp = xpath("$FILE",'/box/source/name');
  $P = getfirst(@temp);
  @temp = xpath("$FILE",'/box/package/post-install');
  $post_install = getfirst(@temp);

  my $binfile = $file;
#  print "file_0: $binfile\n";
  $binfile =~ s/src.box$/*.box/;
#  print "file_a: $binfile\n";
  $binfile =~ s/\Q$P\E-/$P\*-/;
#  print "file_b: $binfile\n";
  my @files = glob($binfile);
  foreach my $f (@files)
  {
#    print "f: $f\n";
    if ($f ne $file)
    {
      $binfile = $f;
    }
  }
#  print "file: $binfile\n";
  open(CMD,"tar xfO$COMPRESS"." $binfile cache/buildinfo 2>/dev/null|");
  my $timestr = <CMD>;
  close(CMD);
  chomp($timestr);

  if (length($timestr) == 0)
  {
    $timestr = format_time(localtime(time));
  }
  my $logdir = "$TESTSDIR/trunk/$timestr" . "-$P";
  my $logfile = "$logdir/$P-$V-$R.log";
  my $shortlogfile = "$timestr" . "-$P/$P-$V-$R.log";
  system "mkdir -p $logdir";
  open(LOGFILE,">$logfile");
  system "cd $TESTSDIR/trunk/; ln -sf $shortlogfile last.log";


  # check for trailling spaces in XML box
  my $line = 0;
  my @trailing = ();
  open(BOX,"$FILE");
  while (<BOX>) {
    $line++;
    if (/^.*([\ \t]+)$/) {
      push @trailing, $line;
    }
  }
  if (@trailing) {
    @trailing = sort @trailing;
    print LOGFILE "------------------------------------------------------------------------------\n";
    print LOGFILE "\n* Trailing characters in XML box file in line(s): @trailing\n";
  }
  close(BOX);



  #FIXME: dynymic arch (?)
  my %new_packages;
  my %factory_packages;
  my %sizeratio;
  my $ARCH = "i686";
  my $binary_file = "$P-$V-$R.$ARCH.box";
  my $devel_file = "$P-devel-$V-$R.$ARCH.box";
  my $static_file = "$P-static-$V-$R.$ARCH.box";
  if (-f "$binary_file") {
    $new_packages{'binary'} = "$binary_file";
  }
  if (-f "$devel_file") {
    $new_packages{'devel'} = "$devel_file";
  }
  if (-f "$static_file") {
    $new_packages{'static'} = "$static_file";
  }
  my @factory_files = glob("$BOXROOT/FACTORY/main/$P/$P*.src.box") ;
  my $nFiles = scalar(@factory_files);
  if ($nFiles == 0) {
    die "\n* There is no source package in the factory main!\n";
  } elsif ($nFiles > 1) {
    die "\n* There are too many source packages in the factory main!\n";
  } elsif ($nFiles == 1) {
    $file = $factory_files[0];
    open(CMD,"tar xf$COMPRESS"." $file $FFILE 2>/dev/null|");
    close(CMD);
    if (! -f $FFILE) {
      die "* Bad file $file!\n";
    }
    system "mv package.box package_factory.box";
    $FFILE = "package_factory.box";
    @temp = xpath("$FFILE",'/box/source/name/@version');
    $fV = getfirst(@temp);
    @temp = xpath("$FFILE",'/box/source/name/@release');
    $fR = getfirst(@temp);
    @temp = xpath("$FFILE",'/box/source/name');
    $fP = getfirst(@temp);
    my $fbinary_file = "$BOXROOT/FACTORY/main/$P/$fP-$fV-$fR.$ARCH.box";
    my $fdevel_file = "$BOXROOT/FACTORY/main/$P/$fP-devel-$fV-$fR.$ARCH.box";
    my $fstatic_file = "$BOXROOT/FACTORY/main/$P/$fP-static-$fV-$fR.$ARCH.box";
    if (-f "$fbinary_file") {
      $factory_packages{'binary'} = "$fbinary_file";
    }
    if (-f "$fdevel_file") {
      $factory_packages{'devel'} = "$fdevel_file";
    }
    if (-f "$fstatic_file") {
      $factory_packages{'static'} = "$fstatic_file";
    }
  }
  # read size of files
  my @sizes = (0,0,0,0,0,0);
  if ($new_packages{'binary'}) {
    $sizes[0] = (stat($new_packages{'binary'}))[7];
  }
  if ($new_packages{'devel'}) {
    $sizes[1] = (stat($new_packages{'devel'}))[7];
  }
  if ($new_packages{'static'}) {
    $sizes[2] = (stat($new_packages{'static'}))[7];
  }
  if ($factory_packages{'binary'}) {
    $sizes[3] = (stat($factory_packages{'binary'}))[7];
  }
  if ($factory_packages{'devel'}) {
    $sizes[4] = (stat($factory_packages{'devel'}))[7];
  }
  if ($factory_packages{'static'}) {
    $sizes[5] = (stat($factory_packages{'static'}))[7];
  }
  my $sum = $sizes[3] + $sizes[4] +  $sizes[5];
  if ($sum == 0) {
    $sizeratio{'all'} = 0;
  } else {
    $sizeratio{'all'} = 100 * ($sizes[0] + $sizes[1] +  $sizes[2]) / $sum;
  }
  # check for binary package
  if ($new_packages{'binary'}) {
    if ($factory_packages{'binary'}) {
       diff_package($factory_packages{'binary'}, $new_packages{'binary'}, \*LOGFILE);
       $sizeratio{'binary'} = 100 * $sizes[0] / $sizes[3];
    } else {
      print LOGFILE "\n* Missing binary package in the factory main!\n";
    }
  } elsif ($factory_packages{'binary'}) {
    print LOGFILE "\n* There is no binary package!\n";
  }

  # check for devel package
  if ($new_packages{'devel'}) {
    if ($factory_packages{'devel'}) {
       diff_package($factory_packages{'devel'}, $new_packages{'devel'}, \*LOGFILE);
       $sizeratio{'devel'} = 100 * $sizes[1] / $sizes[4];;
    } else {
      print LOGFILE "\n* Missing devel package in the factory main!\n";
    }
  } elsif ($factory_packages{'devel'}) {
    print LOGFILE "\n* There is no devel package!\n";
  }

  # check for static package
  if ($new_packages{'static'}) {
    if ($factory_packages{'static'}) {
       diff_package($factory_packages{'static'}, $new_packages{'static'}, \*LOGFILE);
       $sizeratio{'static'} = 100 * $sizes[2] / $sizes[5];;
    } else {
      print LOGFILE "\n* Missing static package in the factory main!\n";
    }
  } elsif ($factory_packages{'static'}) {
    print LOGFILE "\n* There is no static package!\n";
  }
  my @allbinaryfiles = ();
  if ($new_packages{'binary'}) {
    @allbinaryfiles = list_package($new_packages{'binary'});
  }
  my @alldevelfiles = ();
  if ($new_packages{'devel'}) {
    @alldevelfiles = list_package($new_packages{'devel'});
  }
  my @allstaticfiles = ();
  if ($new_packages{'static'}) {
    @allstaticfiles = list_package($new_packages{'static'});
  }
  # factory files
  my @fallbinaryfiles = ();
  if ($factory_packages{'binary'}) {
    @fallbinaryfiles = list_package($factory_packages{'binary'});
  }
  my @falldevelfiles = ();
  if ($factory_packages{'devel'}) {
    @falldevelfiles = list_package($factory_packages{'devel'});
  }
  my @fallstaticfiles = ();
  if ($factory_packages{'static'}) {
    @fallstaticfiles = list_package($factory_packages{'static'});
  }
  # unpack files from all packages needed to further examines (tests)
  $TMPUNPACKED = "$TEMPDIR/unpacked$$";
  $TMPUNPACKEDOLD = "$TEMPDIR/unpackedold$$";
  system "mkdir -p $TMPUNPACKED $TMPUNPACKEDOLD";
  chomp(my $pwd = `pwd`);

  for my $pkg ($new_packages{'binary'},$new_packages{'devel'},$new_packages{'static'})
  {
    next if !($pkg);
    $pkg =~ /(.*)\.box$/;
    my $datafile = "$1.tar.gz";
    my @arr = split(/\//, $datafile);
    $datafile = $arr[$#arr];
    system "cd $TMPUNPACKED; tar xfO $pwd/$pkg data/$datafile | tar xz bin sbin usr/bin usr/sbin lib usr/lib usr/share/applications usr/libexec 2>/dev/null";
    system "cd $TMPUNPACKED; tar xf $pwd/$pkg cache/buildrequires cache/buildprovides cache/provides cache/requires cache/configurehelp cache/news cache/post-install cache/perms 2>/dev/null";
    system "cd $TMPUNPACKED; mkdir -p cache; touch cache/buildrequires cache/buildprovides cache/provides cache/requires cache/configurehelp cache/news cache/post-install cache/perms";
  }
  for my $pkg ($factory_packages{'binary'},$factory_packages{'devel'},$factory_packages{'static'})
  {
    next if !($pkg);
    $pkg =~ /(.*)\.box$/;
    my $datafile = "$1.tar.gz";
    my @arr = split(/\//, $datafile);
    $datafile = $arr[$#arr];
    system "cd $TMPUNPACKEDOLD; tar xfO $pkg data/$datafile | tar xz bin sbin usr/bin usr/sbin lib usr/lib usr/share/applications 2>/dev/null";
    system "cd $TMPUNPACKEDOLD; tar xf $pkg cache/buildrequires cache/buildprovides cache/provides cache/requires cache/configurehelp cache/news cache/post-install cache/perms 2>/dev/null";
    system "cd $TMPUNPACKEDOLD; mkdir -p cache; touch cache/buildrequires cache/buildprovides cache/provides cache/requires cache/configurehelp cache/news cache/post-install cache/perms";
  }
  # Test 1: warning if there desktop files in bad place
  foreach (@allbinaryfiles) {
    chomp;
    if (/usr\/share\/gnome\/apps\/.+\/.+\.desktop/) {
      print LOGFILE "------------------------------------------------------------------------------\n";
      print LOGFILE "\n* Desktop file in bad place '$_' (should be in /usr/share/applications/)\n";
    }
  }
  # Test 2: validate desktop files only (should be only in binary package)
  # also read name and comment from desktop files
  my @desktopfiles = ();
  my @desktopnames = ();
  my %desktopshort;
  my %desktoplong;
  if ($new_packages{'binary'}) {
    foreach (@allbinaryfiles) {
      my $path = $_;
      if (/usr\/share\/applications\/(.+)\.desktop/) {
        my $desktopname = $1;
        push @desktopfiles, $path;
	push @desktopnames, $desktopname;
	open(DSK,"$TMPUNPACKED/$path");
	while  (my $line = <DSK>) {
	  chomp($line);
	  if ($line =~ /^Name\=[\s]*(.+)$/) {
	    $desktopshort{$desktopname} = $1;
	  }
	  if ($line =~ /^Comment\=[\s]*(.+)$/) {
	    $desktoplong{$desktopname} = $1;
	  }
	}
	close(DSK);
      }
    }
    foreach my $desktopfile (@desktopfiles) {
        my $code = system "desktop-file-validate $TMPUNPACKED/$desktopfile &>/dev/null";
 	if ($code != 0) {
          print LOGFILE "------------------------------------------------------------------------------\n";
	  print LOGFILE "\n* File $desktopfile has error(s)!\n";
	}
    }
  }
  # Test 3: check for FHS standard (eg. bad path for documentation, manual)
  # FIXME: add more tests
  my %bad;
  foreach (@allbinaryfiles) {
    chomp;
    if (/usr\/doc\//) {
      $bad{'doc'} = 1;
    }
    if (/usr\/man\//) {
      $bad{'man'} = 1;
    }
  }
  if ($bad{'doc'}) {
      print LOGFILE "------------------------------------------------------------------------------\n";
      print LOGFILE "\n* Bad place for documentation (should be /usr/share/doc/)!\n";
  }
  if ($bad{'man'}) {
      print LOGFILE "------------------------------------------------------------------------------\n";
      print LOGFILE "\n* Bad place for manuals (should be /usr/share/man/)!\n";
  }
  # Test 4: check libraries for suffix so (need to use libtoolize to fix it)
  foreach (@allbinaryfiles) {
    chomp;
    if (/usr\/lib\/[^\.]+\.[0-9]+$/) {
      $bad{'soname'} = 1;
    }
  }
  if ($bad{'soname'}) {
      print LOGFILE "------------------------------------------------------------------------------\n";
      print LOGFILE "\n* Bad suffix for library! (probably need to use libtoolize to fix it)\n";
  }
  # Test 5: mime files exists but thre is no macro for mime-update
#  foreach (@allbinaryfiles) {
#    chomp;
#    if (/usr\/share\/mime\/.+\.xml$/) {
#      $bad{'mime'} = 1;
#    }
#  }
#  if ($bad{'mime'}) {
#    if (!($post_install =~ /\%\{update\-mime\}/)) {
#      print LOGFILE "\n* Mime detected but there is no macro %{update-mime} in post-install!\n";
#    }
#  }
#  if ($post_install =~ /\%\{update\-mime\}/) {
#    if (!($bad{'mime'})) {
#      print LOGFILE "\n* Macro %{update-mime} defined in post-install but not used!\n";
#    }
#  }
  # Test 6: gconf schemas file exists but there is no macro for installing
#  foreach (@allbinaryfiles) {
#    chomp;
#    if (/etc\/gconf\/schemas\/.+\.schemas$/) {
#      $bad{'schemas'} = 1;
#    }
#  }
#  if ($bad{'schemas'}) {
#    if (!($post_install =~ /\%\{gconf_schemas_install\}/)) {
#      print LOGFILE "\n* GConf schemas detected but there is no macro %{gconf_schemas_install} in post-install!\n";
#    }
#  }
#  if (($post_install =~ /\%\{gconf_schemas_install\}/)) {
#    if (!($bad{'schemas'})) {
#      print LOGFILE "\n* Macro %{gconf_schemas_install} defined in post-install but not used!\n";
#    }
#  }
  # Test 7: scrollkeeper documentation exists but thre is no macro for update
#  foreach (@allbinaryfiles) {
#    chomp;
#    if (/usr\/share\/omf\/.+\/.+\.omf$/) {
#      $bad{'scrollkeeper'} = 1;
#    }
#  }
#  if ($bad{'scrollkeeper'}) {
#    if (!($post_install =~ /\%\{scrollkeeper\-update\}/)) {
#      print LOGFILE "\n* Scrollkeeper documentation detected but there is no macro %{scrollkeeper-update} in post-install!\n";
#    }
#  }
#  if (($post_install =~ /\%\{scrollkeeper\-update\}/)) {
#    if (!($bad{'scrollkeeper'})) {
#      print LOGFILE "\n* Macro %{scrollkeeper-update} defined in post-install but not used!\n";
#    }
#  }
  # Test 8: check for pkg-config files and display it as 'provides'
  # also read from .pc files some informations
  my @pkgconfignames = ();
  my %pkgconfigshort;
  my %pkgconfiglong;
  foreach my $path (@allbinaryfiles,@alldevelfiles) {
    chomp($path);
    if ($path =~ /usr\/lib\/pkgconfig\/(.+)\.pc$/) {
      my $pcname = $1;
      push @pkgconfignames, $pcname;
      open(PKG,"$TMPUNPACKED/$path");
      while  (my $line = <PKG>) {
        chomp($line);
	if ($line =~ /^Name\:[\s]*(.+)$/) {
	  $pkgconfigshort{$pcname} = $1;
	}
	if ($line =~ /^Description\:[\s]*(.+)$/) {
	  $pkgconfiglong{$pcname} = $1;
	}
      }
      close(PKG);
    }
  }
#  if (@pkgconfignames) {
#    print LOGFILE "\n* PROVIDES (pkg-config): @pkgconfignames\n";
#  }
  my @pkgconfigprovides = @pkgconfignames;
  # Test 9: propose division package to independent packages
  if ((scalar(@desktopnames) + scalar(@pkgconfignames)) > 1) {
    print LOGFILE "------------------------------------------------------------------------------";
    print LOGFILE "\n* Package could be divided into independent packages (not implemented yet):\n";
    foreach my $el (@desktopnames) {
      print LOGFILE "- $el\n\t". $desktopshort{$el} . "\n\t" . $desktoplong{$el} ."\n";
    }
    foreach my $el (@pkgconfignames) {
      my $newel = $el;
      if ($el =~ /(.+)\-[0-9\.]+$/) {
        $newel = $1;
      }
      print LOGFILE "- $newel\n\t" . $pkgconfigshort{$el} . "\n\t" . $pkgconfiglong{$el} . "\n";
    }
  }
  # Test 10: print dynamic libraries provides
  my %sonames;
  my %fsonames;
  my @changes1 = ();
  my @changes2 = ();
  my @sonamekeys = ();
  my @fsonamekeys = ();
  # provides from pkgconfig detected before
  # zalenoci type="build" rozpatrywa oddzielnie ?!
  foreach my $provide (@pkgconfigprovides) {
    $sonames{"$provide.pc"} = 1;
  }
  # provides from new packages
  foreach my $path (@allbinaryfiles,@alldevelfiles) {
    # mono .NET stuff
    if (($path =~ /usr\/lib\/([^\/]+\.dll)$/) || ($path =~ /usr\/lib\/mono\/gac\/.*\/([^\/]+\.dll)$/)) {
      open(CMD,"mono-find-provides $TMPUNPACKED/$path 2>/dev/null|");
      while (<CMD>) {
        if (/mono\((.+)\)\s\=\s(.+)$/) {
	  my $fullname = "$1.dll $2";
	  $sonames{$fullname} = 1;
        }
      }
      close(CMD);
    }
    if ($path =~ /(usr\/)?lib\/([^\/]*\.so.*)$/) {
	if (-f "$TMPUNPACKED/$path") {
	    my $wassomething = 0;
	    open(CMD,"readelf -d $TMPUNPACKED/$path 2>/dev/null|");
	    while (<CMD>) {
	      if (/soname.+\[(.+)\]/) {
	        $sonames{$1} = 1;
		$wassomething = 1;
	      }
	    }
	    close(CMD);
	    # sometimes library doesn't have "soname" in readelf -d
	    if ($wassomething == 0) {
	      if ($path =~ /(usr\/)?lib\/(lib[^\/]*\.so)$/) {
	        my $lib = $2;
	        open(FILE,"file -L $TMPUNPACKED/$path 2>/dev/null|");
		my $line = <FILE>;
		close(FILE);
		if ($line =~ /shared object/) {
	          $sonames{$lib} = 1;
		}
	      }
	    }
	}
    }
  }
  # factory provides read from XML box file
  my @newprovides = xpath("$FILE",'/box/package/provides/@value');
  foreach (@newprovides) {
    $fsonames{$_} = 1;
  }

  @sonamekeys = sort(keys %sonames);
  @fsonamekeys = sort(keys %fsonames);
  @changes1 = ();
  @changes2 = ();

#  foreach (@sonamekeys) {
#    print "Provides: $_\n";
#  }

  if (@sonamekeys) {
    foreach my $key (@sonamekeys) {
      if (!$fsonames{$key}) {
        push @changes1, $key;
      }
    }
  }
  if (@fsonamekeys) {
    foreach my $key (@fsonamekeys) {
      if (!$sonames{$key}) {
        push @changes2, $key;
      }
    }
  }
#  if (@changes1 || @changes2) {
#      print LOGFILE "\n* PROVIDES (soname) has changed! Please fix it!\n";
#      if (@changes1) {
#        print LOGFILE "- Please add:\n";
#        foreach my $key (sort @changes1) {
#	  my $BUILD='';
#	  if ($key =~ /\.pc$/) {
#	    $BUILD='type="build" ';
#	    #$key =~ s/(.+)\.pc$/$1/;
#	  }
#          print LOGFILE "\t<provides ".$BUILD."value=\"$key\" />\n";
#	}
#      }
#      if (@changes2) {
#        print LOGFILE "- Please remove (propably):\n";
#        foreach my $key (sort @changes2) {
#	  my $BUILD='';
#	  if ($key =~ /\.pc$/) {
#	    $BUILD='type="build" ';
#	  }
#          print LOGFILE "\t<provides ".$BUILD."value=\"$key\" />\n";
#	}
#      }
#  }

  # Test 11: print dynamic libraries requires
  my %sonames2;
  my %fsonames2;
  my %nonstandard;
#  # detect libraries in non standard directories
#  foreach my $path (@allbinaryfiles,@alldevelfiles) {
#    chomp($path);
#    if (($path =~ /(usr\/)?lib\/.*so.*$/) || ($path =~ /(usr\/)?s?bin\/.+$/) ||
#        ($path =~ /usr\/libexec\/.+$/)) {
#	    # if non standard directory and required library is in the
#	    # same directory then don't add this as "required"
#	    my $lib = $path;
#	    if ($lib !~ /(s?bin|usr\/s?bin|lib|usr\/lib|usr\/libexec)\/[^\/]+$/) {
#	      # non standard dir
#	      my $basename = dirname($lib);
#	      $nonstandard{$basename} = 1;
#	      #print "Non standard: $basename\n";
#	    }
#        }
#  }
#  foreach my $path (@allbinaryfiles,@alldevelfiles) {
#    chomp($path);
#    # mono .NET stuff
#    if ($path =~ /\.(exe|dll)$/) {
#      open(CMD,"mono-find-requires $TMPUNPACKED/$path 2>/dev/null|");
#      while (<CMD>) {
#        if (/mono\((.+)\)\s\=\s(.+)$/) {
#	  my $fullname = "$1.dll $2";
#	  # if library in the same dir then don't add this
#	  my $file = dirname($path) . "$1.dll";
#	  my $washere = 0;
#	  foreach my $path2 (@allbinaryfiles,@alldevelfiles) {
#	    chomp($path2);
#	    if ($file =~ /\Q$path2\E$/) {
#	      $washere = 1;
#	    }
#	  }
#	  if ($washere == 0) {
#	    $sonames2{$fullname} = 1;
#	  }
#        }
#      }
#      close(CMD);
#    }
#    # soname stuff
#    if (($path =~ /(usr\/)?lib\/.*so.*$/) || ($path =~ /(usr\/)?s?bin\/.+$/) ||
#        ($path =~ /usr\/libexec\/.+$/)) {
#	if (-f "$TMPUNPACKED/$path") {
#	    open(CMD,"readelf -d $TMPUNPACKED/$path 2>/dev/null|");
#	    while (<CMD>) {
#	      if (/Shared\slibrary.+\[(.+)\]/) {
#	        my $libname = $1;
#		my $pathdirname = dirname($path);
#
#	  my $file = dirname($path) . "$1";
#	  my $washere = 0;
#	  foreach my $path2 (@allbinaryfiles,@alldevelfiles) {
#	    chomp($path2);
#	    if ($file =~ /\Q$path2\E$/) {
#	      $washere = 1;
#	    }
#	  }
#	  if ($washere == 0) {
#	    $sonames2{$libname} = 1;
#	  }
#	      }
#	    }
#	    close(CMD);
#	}
#    }
#  }
  # factory requires read from XML box file
  my @newrequires = xpath("$FILE",'/box/package/requires/@value');
  foreach (@newrequires) {
    $fsonames2{$_} = 1;
  }

  @sonamekeys = sort(keys %sonames2);
  @fsonamekeys = sort(keys %fsonames2);
  @changes1 = ();
  @changes2 = ();

  if (@sonamekeys) {
    foreach my $key (@sonamekeys) {
      if (!$fsonames2{$key}) {
        push @changes1, $key;
      }
    }
  }
  if (@fsonamekeys) {
    foreach my $key (@fsonamekeys) {
      if (!$sonames2{$key}) {
        push @changes2, $key;
      }
    }
  }
#  if (@changes1 || @changes2) {
#      print LOGFILE "\n* REQUIRES (soname) has changed! Please fix it!\n";
#      if (@changes1) {
#        print LOGFILE "- Please add:\n";
#        foreach my $key (sort @changes1) {
#          print LOGFILE "\t<requires value=\"$key\" />\n";
#	}
#      }
#      if (@changes2) {
#        print LOGFILE "- Please remove (propably):\n";
#        foreach my $key (sort @changes2) {
#          print LOGFILE "\t<requires value=\"$key\" />\n";
#	}
#      }
#  }
  # Test12: BUILD PROVIDES

  my %req1;
  my %req2;


  # read requires defined in XML box
  my @boxrequires = xpath("$FILE",'/box/build/requires/@value');
  foreach (@boxrequires) {
    $req2{$_} = 1;
  }
  # read (from cache/buildrequires) requires autodetected
#  if (-f "$TMPUNPACKED/cache/buildrequires") {
#    open(BREQUIRES,"$TMPUNPACKED/cache/buildrequires");
#    while (<BREQUIRES>) {
#      if (!/^\#/) {
#        chomp;
#        $req1{$_} = 1;
#      }
#    }
#    close(BREQUIRES);
#  }

  my @reqkeys1 = sort(keys %req1);
  my @reqkeys2 = sort(keys %req2);
  @changes1 = ();
  @changes2 = ();

  if (@reqkeys1) {
    foreach my $key (@reqkeys1) {
      if (!$req2{$key}) {
        push @changes1, $key;
      }
    }
  }
  if (@reqkeys2) {
    foreach my $key (@reqkeys2) {
      if (!$req1{$key}) {
        push @changes2, $key;
      }
    }
  }
#  if (@changes1 || @changes2) {
#      print LOGFILE "\n* BUILD REQUIRES has changed! Please fix it!\n";
#      if (@changes1) {
#        print LOGFILE "- Please add to build section:\n";
#        foreach my $key (sort @changes1) {
#          print LOGFILE "\t<requires value=\"$key\" />\n";
#	}
#      }
#      if (@changes2) {
#        print LOGFILE "- Please remove from build section (propably):\n";
#        foreach my $key (sort @changes2) {
#          print LOGFILE "\t<requires value=\"$key\" />\n";
#	}
#      }
#  }

  # Test13: Check difference betwean linking programs
  # readelf -d $1 | grep "Shared"| cut -d [ -f2 | cut -d ] -f1 | sort
  my %sharedlibs;
  foreach my $path (@allbinaryfiles,@alldevelfiles) {
    chomp($path);
    if (($path =~ /(usr\/)?lib\/lib.*so.*$/) || ($path =~ /usr\/s?bin\/.+$/) ||
        ($path =~ /s?bin\/.+$/) || ($path =~ /usr\/libexec\/.+$/)) {
	if ((-f "$TMPUNPACKED/$path") && (-f "$TMPUNPACKEDOLD/$path")) {
	    open(CMD,"readelf -d $TMPUNPACKED/$path 2>/dev/null|sort|");
	    open(TMP,">$TMPUNPACKED/so1");
	    while (<CMD>) {
	      if (/Shared\slibrary.+\[(.+)\]/) {
		print TMP "$1\n";
	      }
	    }
	    close(TMP);

	    open(CMD,"readelf -d $TMPUNPACKEDOLD/$path 2>/dev/null|sort|");
	    open(TMP,">$TMPUNPACKED/so2");
	    while (<CMD>) {
	      if (/Shared\slibrary.+\[(.+)\]/) {
		print TMP "$1\n";
	      }
	    }
	    close(TMP);
	    close(CMD);

	    open(DIFF,"diff -u $TMPUNPACKED/so2 $TMPUNPACKED/so1|");
	    if (<DIFF>) {
              print LOGFILE "\n------------------------------------------------------------------------------";
	      print LOGFILE "\n* Difference linking for $path:\n";
	      my $skip = 0;
              while (<DIFF>) {
               chomp;
              if ($skip > 1) {
                print LOGFILE "$_\n";
              }
              $skip++;
            }
	    }
	    close(DIFF);
       }
    }
  }

  # Test 14: Warn when program uses old MIME system
  foreach my $path (@allbinaryfiles,@alldevelfiles) {
    chomp($path);
    if ($path =~ /usr\/share\/mime\-info\/(.+)\.mime$/) {
      $bad{'oldmime'} = 1;
    }
  }
  if ($bad{'oldmime'}) {
    if ($bad{'mime'}) {
      print LOGFILE "------------------------------------------------------------------------------\n";
      print LOGFILE "\n* Application uses old mime-info (not needed) and NEW freedesktop.org\n";
    } else {
      print LOGFILE "------------------------------------------------------------------------------\n";
      print LOGFILE "\n* Application uses OLD mime-info and should be it replaced by new freedesktop.org shared-mime-info\n";
    }
  }
  # Test 15: check file owner/permissions (eg. must be executable for libraries)
  my @nonexec = ();
  foreach my $path (@allbinaryfiles,@alldevelfiles) {
    chomp($path);
    if (($path =~ /(usr\/)?lib\/lib.*so.+\..+$/) || ($path =~ /usr\/libexec\/.+$/)) {
	if ((-f "$TMPUNPACKED/$path") && !(-x "$TMPUNPACKED/$path")) {
	  push @nonexec, $path;
	}
    }
  }
  if (@nonexec) {
    print LOGFILE "------------------------------------------------------------------------------\n";
    print LOGFILE "\n* Libraries should be executable: @nonexec\n";
  }
  # Test 16: show difference betwean configurehelp
  open(DIFF,"diff -u $TMPUNPACKEDOLD/cache/configurehelp $TMPUNPACKED/cache/configurehelp|");
  if (<DIFF>) {
    print LOGFILE "\n------------------------------------------------------------------------------";
    print LOGFILE "\n* Difference in configure help:\n";
    my $skip = 0;
    while (<DIFF>) {
      chomp;
      if ($skip > 1) {
        print LOGFILE "$_\n";
      }
      $skip++;
    }
  }
  
  close(DIFF);
  # Test 17: show difference betwean buildrequires
  open(DIFF,"diff -u $TMPUNPACKEDOLD/cache/buildrequires $TMPUNPACKED/cache/buildrequires|");
  if (<DIFF>) {
    print LOGFILE "\n------------------------------------------------------------------------------";
    print LOGFILE "\n* Difference in buildrequires:\n";
    my $skip = 0;
    while (<DIFF>) {
      chomp;
      if ($skip > 1) {
        print LOGFILE "$_\n";
      }
      $skip++;
    }
  }
  close(DIFF);
  # Test 18: show difference betwean buildprovides
  open(DIFF,"diff -u $TMPUNPACKEDOLD/cache/buildprovides $TMPUNPACKED/cache/buildprovides|");
  if (<DIFF>) {
    print LOGFILE "\n------------------------------------------------------------------------------";
    print LOGFILE "\n* Difference in buildprovides:\n";
    my $skip = 0;
    while (<DIFF>) {
      chomp;
      if ($skip > 1) {
        print LOGFILE "$_\n";
      }
      $skip++;
    }
  }
  close(DIFF);
  # Test 19: show difference betwean requires
  open(DIFF,"diff -u $TMPUNPACKEDOLD/cache/requires $TMPUNPACKED/cache/requires|");
  if (<DIFF>) {
    print LOGFILE "\n------------------------------------------------------------------------------";
    print LOGFILE "\n* Difference in requires:\n";
    my $skip = 0;
    while (<DIFF>) {
      chomp;
      if ($skip > 1) {
        print LOGFILE "$_\n";
      }
      $skip++;
    }
  }
  # Test 20: show difference betwean provides
  open(DIFF,"diff -u $TMPUNPACKEDOLD/cache/provides $TMPUNPACKED/cache/provides|");
  if (<DIFF>) {
    print LOGFILE "\n------------------------------------------------------------------------------";
    print LOGFILE "\n* Difference in provides:\n";
    my $skip = 0;
    while (<DIFF>) {
      chomp;
      if ($skip > 1) {
        print LOGFILE "$_\n";
      }
      $skip++;
    }
  }
  close(DIFF);
  # Test 21: show difference betwean post-install
  open(DIFF,"diff -u $TMPUNPACKEDOLD/cache/post-install $TMPUNPACKED/cache/post-install|");
  if (<DIFF>) {
    print LOGFILE "\n------------------------------------------------------------------------------";
    print LOGFILE "\n* Difference in post-install:\n";
    my $skip = 0;
    while (<DIFF>) {
      chomp;
      if ($skip > 1) {
        print LOGFILE "$_\n";
      }
      $skip++;
    }
  }
  close(DIFF);
  # Test 22: show difference betwean perms
  open(DIFF,"diff -u $TMPUNPACKEDOLD/cache/perms $TMPUNPACKED/cache/perms|");
  if (<DIFF>) {
    print LOGFILE "\n------------------------------------------------------------------------------";
    print LOGFILE "\n* Difference in perms:\n";
    my $skip = 0;
    while (<DIFF>) {
      chomp;
      if ($skip > 1) {
        print LOGFILE "$_\n";
      }
      $skip++;
    }
  }
  close(DIFF);
  # Test 23: show difference betwean news
  open(DIFF,"diff -u $TMPUNPACKEDOLD/cache/news $TMPUNPACKED/cache/news|");
  my $endline = 0;
  if (<DIFF>) {
    my $skip = 0;
    my $line;
    while (<DIFF>) {
      chomp;
      $line = substr($_,0,1);
      if ($skip > 1) {
        if ($line ne '-') {
	  $endline = $skip;
	}
      }
      $skip++;
    }
  }
  close(DIFF);
  open(DIFF,"diff -u $TMPUNPACKEDOLD/cache/news $TMPUNPACKED/cache/news|");
  if (<DIFF>) {
    print LOGFILE "\n------------------------------------------------------------------------------";
    print LOGFILE "\n* Difference in news:\n";
    my $skip = 0;
    while (<DIFF>) {
      chomp;
      if ($skip > 1 && $skip <= $endline) {
        print LOGFILE "$_\n";
      }
      $skip++;
    }
  }
  close(DIFF);
  # Test XX: check packages size
  print LOGFILE "\n------------------------------------------------------------------------------";
  print LOGFILE "\n* Size ratio:\n";
  if ($sizeratio{'binary'}) {
    print LOGFILE "- binary  " . sprintf("%0.2f", $sizeratio{'binary'}) . "%\n";
  }
  if ($sizeratio{'devel'}) {
    print LOGFILE "- devel   " . sprintf("%0.2f", $sizeratio{'devel'}) . "%\n";
  }
  if ($sizeratio{'static'}) {
    print LOGFILE "- static  " . sprintf("%0.2f", $sizeratio{'static'}) . "%\n";
  }
  if ($sizeratio{'all'}) {
    print LOGFILE "- overall " . sprintf("%0.2f", $sizeratio{'all'}) . "%\n";
  }

  # TODO: add more inteligent tests :-)

  # clean temporary files
  close(LOGFILE);
  logs_management();
  system("rm -f $FILE");
  system("rm -rf $TMPUNPACKED $TMPUNPACKEDOLD");
  system("rm -f $FILE");
  system("rm -f $FFILE");
}

sub insert_deps
{
  my $file = $_[0];
  my $dbh = $_[1];
  my $dirname = $_[2];
  my $GROUP = $_[3];
  my $sth;
#    die($dirname);
  my @e = ();
  my %pkg_provides;
  my %pkg_provides_db;
#  my %pkg_requires_pkg; gnome-desktop -> package.scrollkeeper ma zabronic bo nie jest zdefiniowany
#  my %pkg_requires_db;

      open(G, "box -show provides $file 2>/dev/null|");
      while (<G>) {
	@e = split(/\s+/);
        chomp;
	$pkg_provides{$e[0]} = 1;
      }
      close(G);
      
#      open(G, "box -show buildprovides $file 2>/dev/null|");
#      while (<G>) {
#	@e = split(/\s+/);
#        chomp;
#	$pkg_provides{$e[0]} = 1;
#      }
#      close(G);

  $sth = $dbh->prepare("select p.name from provides p, packages pkg where
  pkg.id = p.package_id and pkg.name = '$dirname' and p.type = 0");
  $sth->execute();
  while (my @row = $sth->fetchrow_array)
  {
    #printf "%s\n", $row[0];
    $pkg_provides_db{$row[0]} = 1;
  }
  
  my $error = 0;
  foreach my $key (keys %pkg_provides_db) {
    if (! $pkg_provides{$key}) {
	$sth = $dbh->prepare("select pkg.name from requires r, packages pkg where
	r.name = '$key' and pkg.id = r.package_id and pkg.name != '$dirname'");
	$sth->execute();
	while (my @row = $sth->fetchrow_array)
	{
	    printf "package '%s' depends on %s\n", $row[0], $key;
	    $error = 1;
	}
    }
  }

  if ($error == 1) {
    die("Please correct above packages!");
  }

    $sth = $dbh->prepare("SELECT id FROM packages WHERE name = '$dirname'");
    $sth->execute();
    my @ids = $sth->fetchrow_array;
    my $package_id;
    my $groupnumber = -1;
    if ($GROUP eq 'core') {
      $groupnumber = 0;
    } elsif ($GROUP eq 'extras') {
      $groupnumber = 1;
    }
    if (!@ids) {
        $package_id = $ids[0];
        $dbh->do("INSERT INTO packages VALUES(NULL, '$dirname', '$groupnumber')");
	print "Adding new package: '$dirname', group $groupnumber\n";
    } else {
        $dbh->do("UPDATE packages SET grp = '$groupnumber' WHERE name = '$dirname'");
    }
    # read again package id
    $sth = $dbh->prepare("SELECT id FROM packages WHERE name = '$dirname'");
    $sth->execute();
    @ids = $sth->fetchrow_array;
    $package_id = $ids[0];
      open(G, "box -show provides $file 2>/dev/null|");
      $dbh->do("DELETE FROM provides WHERE package_id = '$package_id' AND type = 0");
      while (<G>) {
	@e = split(/\s+/);
        chomp;
        $dbh->do("INSERT INTO provides VALUES(NULL, $package_id, '$e[0]' , 0)");
	print "Inserting provides for $dirname: '$e[0]'\n";
      }
      close(G);
      open(G, "box -show requires $file 2>/dev/null|");
      $dbh->do("DELETE FROM requires WHERE package_id = '$package_id' AND type = 0");
      while (<G>) {
	@e = split(/\s+/);
        chomp;
        $dbh->do("INSERT INTO requires VALUES(NULL, $package_id, '$e[0]' , 0)");
	print "Inserting requires for $dirname: '$e[0]'\n";
      }
      close(G);
      open(G, "box -show buildprovides $file 2>/dev/null|");
      $dbh->do("DELETE FROM provides WHERE package_id = '$package_id' AND type = 1");
      while (<G>) {
	@e = split(/\s+/);
        chomp;
        $dbh->do("INSERT INTO provides VALUES(NULL, $package_id, '$e[0]' , 1)");
	print "Inserting buildprovides for $dirname: '$e[0]'\n";
      }
      close(G);
      open(G, "box -show buildrequires $file 2>/dev/null|");
      $dbh->do("DELETE FROM requires WHERE package_id = '$package_id' AND type = 1");
      while (<G>) {
	@e = split(/\s+/);
        chomp;
        $dbh->do("INSERT INTO requires VALUES(NULL, $package_id, '$e[0]' , 1)");
	print "Inserting buildrequires for $dirname: '$e[0]'\n";
      }
      close(G);
}

# ============================================================================
# accept_factory(source_package)
# ============================================================================
sub accept_factory
{
  my $file = $_[0];
  my $origfile = $file;
  my @temp = ();
  my $V;
  my $P;
  my $R;
  my $fV;
  my $fP;
  my $fR;
  my $GROUP;
  my $fGROUP;

  my $FILE = "package.box";
  open(CMD,"tar xf$COMPRESS"." $file $FILE 2>/dev/null|");
  close(CMD);
  if (! -f $FILE) {
    die "Bad file!\n";
  }
  @temp = xpath("$FILE",'/box/source/name/@version');
  $V = getfirst(@temp);
  @temp = xpath("$FILE",'/box/source/name/@release');
  $R = getfirst(@temp);
  @temp = xpath("$FILE",'/box/source/name');
  $P = getfirst(@temp);
  @temp = xpath("$FILE",'/box/@group');
  $GROUP = getfirst(@temp);
  #FIXME: dynamic arch (?) not now!
  my %new_packages;
  my %factory_packages;
  my $ARCH = "i686";
  my $binary_file = "$P-$V-$R.$ARCH.box";
  my $devel_file = "$P-devel-$V-$R.$ARCH.box";
  my $static_file = "$P-static-$V-$R.$ARCH.box";
  my $source_file = $origfile;
  if (-f "$binary_file") {
    $new_packages{'binary'} = "$binary_file";
  }
  if (-f "$devel_file") {
    $new_packages{'devel'} = "$devel_file";
  }
  if (-f "$static_file") {
    $new_packages{'static'} = "$static_file";
  }
  my @factory_files = glob("$BOXROOT/FACTORY/main/$P/$P*.src.box") ;
  my $nFiles = scalar(@factory_files);
  my $new = 0;
  if ($nFiles == 0) {
    #die "There is no source package in the factory!\n";
    print STDERR "Warning: Accepting new package\n";
    $new = 1;
  } elsif ($nFiles > 1) {
    die "There are too many source packages in the factory!\n";
  } elsif ($nFiles == 1) {
    $file = $factory_files[0];
    open(CMD,"tar xf$COMPRESS"." $file $FILE 2>/dev/null|");
    close(CMD);
    if (! -f $FILE) {
      die "Bad file!\n";
    }
    @temp = xpath("$FILE",'/box/source/name/@version');
    $fV = getfirst(@temp);
    @temp = xpath("$FILE",'/box/source/name/@release');
    $fR = getfirst(@temp);
    @temp = xpath("$FILE",'/box/source/name');
    $fP = getfirst(@temp);
    @temp = xpath("$FILE",'/box/@group');
    $fGROUP = getfirst(@temp);
    my $fbinary_file = "$BOXROOT/FACTORY/main/$P/$fP-$fV-$fR.$ARCH.box";
    my $fdevel_file = "$BOXROOT/FACTORY/main/$P/$fP-devel-$fV-$fR.$ARCH.box";
    my $fstatic_file = "$BOXROOT/FACTORY/main/$P/$fP-static-$fV-$fR.$ARCH.box";
    my $fsource_file = "$BOXROOT/FACTORY/main/$P/$fP-$fV-$fR.src.box";
    if (-f "$fbinary_file") {
      $factory_packages{'binary'} = "$fbinary_file";
    }
    if (-f "$fdevel_file") {
      $factory_packages{'devel'} = "$fdevel_file";
    }
    if (-f "$fstatic_file") {
      $factory_packages{'static'} = "$fstatic_file";
    }
    $factory_packages{'source'} = "$fsource_file";
    system("rm -f $FILE");
  }
  my $TMPUNPACKED = "$TEMPDIR/unpacked$$";
  system "mkdir -p $TMPUNPACKED";
  chomp(my $pwd = `pwd`);
  # read start building time from cache
  my $STARTBUILD = '';
  my $found = 0;
  for my $pkg ($new_packages{'binary'},$new_packages{'devel'},$new_packages{'static'})
  {
    next if !($pkg);
    next if ($found == 1);
    $pkg =~ /(.*)\.box$/;
    system "cd $TMPUNPACKED; tar xf $pwd/$pkg cache/buildinfo 2>/dev/null";
    if (-f "$TMPUNPACKED/cache/buildinfo" ) {
      $found = 1;
      open(BI,"$TMPUNPACKED/cache/buildinfo");
      my @lines = <BI>;
      chomp($STARTBUILD = $lines[0]);
      close(BI);
    }
  }
  my $logdir="$LOGDIR/trunk/$STARTBUILD-$P";
  my $testsdir="$TESTSDIR/trunk/$STARTBUILD-$P";
  if (! -d "$testsdir")
  {
    # clean temporary files
    system("rm -f $FILE");
    system("rm -rf $TMPUNPACKED");
    if ($new == 1)
    {
      #print STDERR "Please make tests (box -check) after accepting new package!\n";
    } else
    {
      die "You should make tests (box -check) before accepting!!!\n";
    }
  }
  
  use DBI;
  my $dbh;
  my $sth;
  if (! -f "$BOXROOT/FACTORY/latest/database")
  {
  system "mkdir -p $BOXROOT/FACTORY/latest";
  $dbh = DBI->connect("DBI:SQLite:dbname=$BOXROOT/FACTORY/latest/database",'','', {'RaiseError' => 1}) or die;  print "creating\n";
  $dbh->do("CREATE TABLE packages(id INTEGER PRIMARY KEY,
                                  name TEXT,
				  grp SMALLINT)");
  $dbh->do("CREATE TABLE provides(id INTEGER PRIMARY KEY,
                                  package_id INTEGER,
                                  name TEXT,
				  type SMALLINT)");
  $dbh->do("CREATE TABLE requires(id INTEGER PRIMARY KEY,
                                  package_id INTEGER,
                                  name TEXT,
				  type SMALLINT)");
  }
  $dbh = DBI->connect("DBI:SQLite:dbname=$BOXROOT/FACTORY/latest/database",'','', {'RaiseError' => 1}) or die;
  
  {
    if (-f $binary_file) {
      insert_deps($binary_file, $dbh, $P, $GROUP);
    }
    if (-f $devel_file) {
      insert_deps($devel_file, $dbh, $P, $GROUP);
    }
  }

  # clean the factory-old
  system "mkdir -p $BOXROOT/FACTORY/main_backup/$P";
  system "rm -f $BOXROOT/FACTORY/main_backup/$P/*";
  system "mkdir -p $LOGDIR/main/$P";
  system "mkdir -p $LOGDIR/main_backup/$P";
  system "rm -f $LOGDIR/main_backup/$P/*";
  system "mkdir -p $TESTSDIR/main/$P";
  system "mkdir -p $TESTSDIR/main_backup/$P";
  system "rm -f $TESTSDIR/main_backup/$P/*";
  system "mkdir -p $BOXROOT/FACTORY/main/$P/";
  system "mkdir -p $BOXROOT/FACTORY/latest/core/{bin,dev,sta,src}";
  system "mkdir -p $BOXROOT/FACTORY/latest/extras/{bin,dev,sta,src}";
  system "mkdir -p $BOXROOT/FACTORY/latest/candidate/{bin,dev,sta,src}";
  
  # move files from the factory main to the factory main_backup
  if ($new == 1) {
  } else {
    system "mv $BOXROOT/FACTORY/main/$P/* $BOXROOT/FACTORY/main_backup/$P/ &>/dev/null";
    system "mv $LOGDIR/main/$P/* $LOGDIR/main_backup/$P/ &>/dev/null";
    system "mv $LOGDIR/main/$P/* $LOGDIR/main_backup/$P/ &>/dev/null";
    system "mv $TESTSDIR/main/$P/* $TESTSDIR/main_backup/$P/ &>/dev/null";
    system "mv $TESTSDIR/main/$P/* $TESTSDIR/main_backup/$P/ &>/dev/null";
  }
  # move files from main to the factory
  if ($new == 1) {
    # origfile == source package
    #system "cp $origfile $BOXROOT/FACTORY/main/$P/";
    system "mv $origfile $BOXROOT/FACTORY/main/$P/";
  } else 
  {
    system "mv $origfile $BOXROOT/FACTORY/main/$P/";
    system "cp $logdir/* $LOGDIR/main/$P/";
    system "cp $testsdir/* $TESTSDIR/main/$P/";
  }
  # call logs_management
  logs_management();
  my $cmd = "mv";
  if ($new == 1) {
    #$cmd = "cp";
  }
  if ($factory_packages{'binary'}) {
    system "rm -f $BOXROOT/FACTORY/latest/$fGROUP/bin/" . basename($factory_packages{'binary'});
  }
  if ($factory_packages{'devel'}) {
    system "rm -f $BOXROOT/FACTORY/latest/$fGROUP/dev/" . basename($factory_packages{'devel'});
  }
  if ($factory_packages{'static'}) {
    system "rm -f $BOXROOT/FACTORY/latest/$fGROUP/sta/" . basename($factory_packages{'static'});
  }
  if ($factory_packages{'source'}) {
	system "rm -f $BOXROOT/FACTORY/latest/$fGROUP/src/" . basename($factory_packages{'source'});
  }
  if ($new_packages{'binary'}) {
    system "$cmd " . $new_packages{'binary'} . " $BOXROOT/FACTORY/main/$P/";
    system "ln -sf $BOXROOT/FACTORY/main/$P/$binary_file $BOXROOT/FACTORY/latest/$GROUP/bin/";
  }
  if ($new_packages{'devel'}) {
    system "$cmd " . $new_packages{'devel'} . " $BOXROOT/FACTORY/main/$P/";
    system "ln -sf $BOXROOT/FACTORY/main/$P/$devel_file $BOXROOT/FACTORY/latest/$GROUP/dev/";
  }
  if ($new_packages{'static'}) {
    system "$cmd " . $new_packages{'static'} . " $BOXROOT/FACTORY/main/$P/";
    system "ln -sf $BOXROOT/FACTORY/main/$P/$static_file $BOXROOT/FACTORY/latest/$GROUP/sta/";
  }
  system "ln -sf $BOXROOT/FACTORY/main/$P/$source_file $BOXROOT/FACTORY/latest/$GROUP/src/";
  # clean temporary files
  system("rm -f $FILE");
  system("rm -rf $TMPUNPACKED");
}


# ============================================================================
# check_dependency($type, @groups)
# ============================================================================
sub check_dependency
{
  my $type = shift;
  my @groups = @_;

  use DBI;
  my $dbh;
  my $sth;
  my $sql_type;
  my $sql_grp;
  
  if ($type eq "install") {
    $sql_type = 0;
  } elsif ($type eq "build") {
    $sql_type = 1;
  }
  
  my $group_bits = 0;
 
  foreach (@groups) {
    if ($_ eq "core") {
	$group_bits = $group_bits | 1;
    }
    if ($_ eq "extras") {
	$group_bits = $group_bits | 2;
    }
  }
  $group_bits--;
  if ($group_bits == 0) {
    $sql_grp = "= 0";
  }
  if ($group_bits == 1) {
    $sql_grp = "= 1";
  }
  if ($group_bits == 2) {
    $sql_grp = ">= 0";
  }
  

  
  print "$group_bits\n";

  $dbh = DBI->connect("DBI:SQLite:dbname=$BOXROOT/FACTORY/latest/database",'','', {'RaiseError' => 1}) or die;
  $sth = $dbh->prepare("select p.name, r.name from requires r, packages p where r.name in
  (select ir.name from requires ir, packages ip where ir.type = $sql_type and
  ir.package_id = ip.id and ip.grp $sql_grp except select iq.name from provides iq,
  packages ip where iq.type = $sql_type and iq.package_id = ip.id and ip.grp $sql_grp) and
  p.id = r.package_id and p.grp $sql_grp order by p.name");
  $sth->execute();
  while (my @row = $sth->fetchrow_array)
  {
    printf "%s\t%s\n", $row[0], $row[1];
  }  
=pod
  my %allprovides;
  my %allrequires;

  print STDERR "Writing deps_big.dot file...\n";
  open(DOT,">deps_big.dot");
  print DOT "digraph deps {\n";
  foreach my $box (glob("$BOXES/*.box")) {
    my $group = getfirst(xpath("$box",'/box/@group'));
    if (grep(/$group/, @groups)) {
      my $package_name = getfirst(xpath("$box",'/box/source/name'));
      my @provides;
      if ($type eq "package") {
        @provides = xpath("$box",'/box/package/provides[@type != "build"]/@value');
      } elsif ($type eq "build") {
        @provides = xpath("$box",'/box/package/provides[@type = "build"]/@value');
      }

      foreach (@provides) {
        s/(.+)\.pc/$1/;
        my @el;
        @el = split(/[\s]/);
        $allprovides{$el[0]} = "$package_name";
	print DOT "\t\"$el[0]\" -> \"pkg_$package_name\"\;\n";
      }
      my @requires = xpath("$box","/box/$type/requires/\@value");
      foreach (@requires) {
        my @el;
        @el = split(/[\s]/);
        $allrequires{$el[0]} = "$package_name";
	print DOT "\t\"pkg_$package_name\" -> \"$el[0]\"\;\n";
      }
    }
  }
  print DOT "}\n";
  close(DOT);

  my %provides_missing;
  my %provides_unused;

  print STDERR "Writing deps.txt file...\n";
  open(TXT,">deps.txt");

  print TXT "Missing provides:\n";
  foreach my $key (sort keys %allrequires) {
    if (! exists $allprovides{$key}) {
      print TXT "$key \t(eg. " . $allrequires{$key} . ")\n";
    }
  }

  print TXT "\n\nUnused provides:\n";
  foreach my $key (sort keys %allprovides) {
    if (! exists $allrequires{$key}) {
      print TXT "$key \t(" . $allprovides{$key} . ")\n";
    }
  }
  close(TXT);

  print STDERR "Writing deps.dot file...\n";
  open(DOT2,">deps.dot");
  print DOT2 "digraph ${type}deps {\n";
  my $xxx = 0;
  foreach my $box (glob("$BOXES/*.box")) {
    my $group = getfirst(xpath("$box",'/box/@group'));
    my %buildreq;
    if (grep(/$group/, @groups)) {
      my $package_name = getfirst(xpath("$box",'/box/source/name'));
      if (length($package_name) == 0) {
        print STDERR "ERROR: $box\n";
      }
#      $xxx++;
#      print STDERR "$xxx\t$package_name\n";
      my @requires = xpath("$box","/box/$type/requires/\@value");
      my @packages = ();
      foreach (@requires) {
        my @el = split(/[\s]/);
	my $element = $el[0];
	my $value = '';
	if ($allprovides{$element}) {
	  $value = $allprovides{$element};
	};
	if (length($value) > 0)
	{
	  push @packages, $value;
	}
      }
	if (!(@packages))
	{
	  print "$package_name\tnone\n";
	}
	my $oldpkg = '';
        foreach my $pkg (sort @packages) {
    	    if (("$oldpkg" ne "$pkg") && ("$pkg" ne "$package_name"))
	     {
	      print "$package_name\t$pkg\n";
	      print DOT2 "\t\"$package_name\" -> \"$pkg\"\;\n";
	    }
	    $oldpkg = $pkg;
      }
    }
  }
  close(DOT2);
=cut
}

# ============================================================================
# check_files(@files)
# ============================================================================
sub check_files
{
  my @groups = @_;
  my %files;
  # dla kadego katalogu w fabryce o ile istniej pliki binarne i nale do odpowiedniej grupy
  # wywietl list  i dla kadego pliku zlicz liczb wystpie
  # list plikw dopisz do pliku files.txt
  open(FILE,">files.txt");
  foreach my $box (glob("$BOXROOT/BOXES/factory/*/*")) {
    if ($box !~ /src.box$/) {
      system "tar xf $box package.box";
      my $group = getfirst(xpath("package.box",'/box/@group'));
      print STDERR "$box...\n";
      if (grep(/$group/, @groups)) {
        my $package_name = getfirst(xpath("package.box",'/box/source/name'));
	my @filelist = list_package($box);
	print FILE "\@$package_name\n";
	foreach my $line (@filelist) {
	  chomp($line);
	  if ($line !~ /\/$/) {
	    $files{$line}++;
	    print FILE "$line\n";
	  }
	}
      }
      system "rm -f package.box";
    }
  }
  close(FILE);
  foreach my $key (sort keys %files) {
    if ($files{$key} > 1) {
      print "$key\t" . $files{$key} . "\n";
    }
  }
}

# ============================================================================
# tidy_box(xml_box)
# ============================================================================
sub tidy_box
{
  print STDERR "Sorry, not implemented yet!\n";
}

# ============================================================================
# generate_report()
# ============================================================================
sub generate_report
{
  system "mkdir -p report";
  open(IDX,">report/index.html");

  print IDX q{
<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01//EN" "http://www.w3.org/TR/html4/strict.dtd">
<html>
<head>
<meta name="author" content="BOX package management system">
<title>Pingwinek GNU/Linux</title>
</head>
<body>
<div align="center"><h1>System statistics</h1></div>
<div align="center"><h3>for Pingwinek GNU/Linux</h3></div>
<table width="100%" border="1">
	<tr>
		<td colspan="1" rowspan="2" width="20%" bgcolor="#BFBFBF" align="center"><b>Package type:</b></td>
		<td colspan="4" align="center" bgcolor="#BFBFBF"><b>Group</b></td>
		<td colspan="2" rowspan="2" align="center" bgcolor="#BFBFBF"><i>Summary:</i></td>
	</tr>
	<tr>
		<td colspan="2" align="center" bgcolor="#E5E5E5"><b>CORE</b></td>
		<td colspan="2" align="center" bgcolor="#E5E5E5"><b>EXTRAS</b></td>
	</tr>
	<tr>
		<td bgcolor="#E5E5E5">Source</td>
};
  print IDX "\t\t<td align=\"center\" bgcolor=\"\#E6E6FA\">500</td>\n";
  print IDX "\t\t<td align=\"center\" bgcolor=\"\#E6E6FA\">700MB</td>\n";
  print IDX "\t\t<td align=\"center\" bgcolor=\"\#E6E6FA\">100</td>\n";
  print IDX "\t\t<td align=\"center\" bgcolor=\"\#E6E6FA\">150MB</td>\n";
  print IDX "\t\t<td align=\"center\" bgcolor=\"\#F0E6F0\">600</td>\n";
  print IDX "\t\t<td align=\"center\" bgcolor=\"\#F0E6F0\">850MB</td>\n";
  print IDX q{
	</tr>
	<tr>
		<td bgcolor="#E5E5E5">Binary</td>
};
#		<td align="center" bgcolor="#E6E6FA">450</td>
#		<td align="center" bgcolor="#E6E6FA">100MB</td>
#		<td align="center" bgcolor="#E6E6FA">100</td>
#		<td align="center" bgcolor="#E6E6FA">150MB</td>
#		<td align="center" bgcolor="#F0E6F0"></td>
#		<td align="center" bgcolor="#F0E6F0"></td>
  print IDX q{
	</tr>
	<tr>
		<td bgcolor="#E5E5E5">Development</td>
};
#		<td align="center" bgcolor="#E6E6FA"></td>
#		<td align="center" bgcolor="#E6E6FA"></td>
#		<td align="center" bgcolor="#E6E6FA"></td>
#		<td align="center" bgcolor="#E6E6FA"></td>
#		<td align="center" bgcolor="#F0E6F0"></td>
#		<td align="center" bgcolor="#F0E6F0"></td>
  print IDX q{
	</tr>
	<tr>
		<td bgcolor="#E5E5E5">Static</td>
};
#		<td align="center" bgcolor="#E6E6FA"></td>
#		<td align="center" bgcolor="#E6E6FA"></td>
#		<td align="center" bgcolor="#E6E6FA"></td>
#		<td align="center" bgcolor="#E6E6FA"></td>
#		<td align="center" bgcolor="#F0E6F0"></td>
#		<td align="center" bgcolor="#F0E6F0"></td>
  print IDX q{
	</tr>
	<tr>
		<td bgcolor="#E5E5E5"><i>Summary:</i></td>
};
#		<td align="center" bgcolor="#F0E6F0">950</td>
#		<td align="center" bgcolor="#F0E6F0"></td>
#		<td align="center" bgcolor="#F0E6F0"></td>
#		<td align="center" bgcolor="#F0E6F0"></td>
#		<td align="center" bgcolor="#FFD0CE">xxx</td>
#		<td align="center" bgcolor="#FFD0CE">100MB</td>
  print IDX q{
	</tr>
	<tr><td colspan="7">&nbsp;</td></tr>
	<tr>
		<td align="right" bgcolor="#BFBFBF"><b>Id:</b></td>
		<td colspan="2" align="center" bgcolor="#BFBFBF"><b>CORE</b></td>
		<td colspan="2" align="center" bgcolor="#BFBFBF"><b>EXTRAS</b></td>
	</tr>
	<tr>
		<td align="right" bgcolor="#E5E5E5">1.</td>
		<td colspan="2" align="center"><a href="">aalib.box</a></td>
		<td colspan="2" align="center"><a href="">nvidia.box</a></td>
	</tr>

</table>

</body>
</html>
};
  close(IDX);
}

# ============================================================================
# print_newsoft()
# ============================================================================

sub print_newsoft
{
  my $mask = '*.box';
  my ($prefix) = (@_);
  if (!defined($prefix))
  {
    $prefix = $config{'BOXES'};
  }

  # yes, yes password is here ;)
  my $dbh = DBI->connect("DBI:mysql:database=pingwinek:host=localhost",'root','toor', {'RaiseError' => 1}) or die;

  print STDERR "Detecting new versions of software. Please wait...\n";

# subfunction
sub read_box
{
  my $SERVER = "";
  my $URL = "";
  my $LATEST = "";
  my $file = shift;
  my $firstpath;
  my @temp;
  # FIXME: read version from DB sources
  @temp = xpath("$file",'/box/source/name/@version');
  my $V = getfirst(@temp);
  # FIXME: read url from DB sources (should be added)
  @temp = xpath("$file",'/box/source[@id="main"]/url');
  $URL = getfirst(@temp);
  # FIXME: name is from DB sources
  @temp = xpath("$file",'/box/source/name');
  my $name = getfirst(@temp);

  my @path;
  my $i=0;
  @path = split(/\//,$URL);
  # STAGE 1
  foreach my $element (@path)
  {
    $i++;
    if ($i<=$#path)
    {
      if ($element eq '%{GNOME.ORG}')
      {
        $LATEST='ftp://ftp.gnome.org/pub/GNOME/sources/';
        $SERVER='GNOME.ORG';
      } elsif ($element eq '%{GNU.ORG}')
      {
        $LATEST='ftp://ftp.gnu.org/pub/gnu/';
        $SERVER='GNU.ORG';
      } elsif ($element eq '%{SF.NET}')
      {
        $LATEST='http://dl.sourceforge.net/';
        $SERVER='SF.NET';
        my @elem = split(/\//, $URL);
        $firstpath = $elem[1];
        $URL =~ s/\%\{SF\.NET\}//;
        $URL =~ s/\./\\\./g;
        $URL =~ s/\+/\\+/g;
        $URL =~ s/\//\\\//g;
        #$URL =~ s/\-/\\\-/g;
      } elsif ($element eq '%{main_version}')
      {
        my $main_version = "";
        my @numbers = split('\.', $V);
        if (scalar(@numbers) >= 2)
        {
          $main_version = "$numbers[0].$numbers[1]";
          $LATEST=$LATEST.$main_version."/";
        } else
        {
          $main_version = "ERROR";
        }
      } else
      {
        $LATEST=$LATEST.$element."/";
      }
    }
      # generic url modifications
      $URL =~ s/\%\{version}/\(\[0\-9\]\[\^\\\>\"\]\+\)/;
  }
  return ($SERVER, $firstpath, $URL, $name, $V, $LATEST);
}
# end of read_box

# subfunction
sub insert_row
{
  my ($dbh, $program, $version, $newversion) = (@_);

  my $sth = $dbh->prepare("SELECT COUNT(*) FROM newsoft WHERE program = '$program' AND version = '$newversion'");
  $sth->execute();
  my @r = $sth->fetchrow_array;
  if ($r[0] == 0)
  {
    my $sth2 = $dbh->do("INSERT INTO newsoft (program, version, date) VALUES('$program', '$newversion', now())");
  }
  $sth->finish();

  # insert to database
  print "$program\t$version\t$newversion\n";
}
# end of insert_row

  # connect to server GNOME.ORG (FIXME: only if were used)
  my $gnomeftp;
  $gnomeftp = Net::FTP->new("ftp.gnome.org", Debug => 0, Passive => 1)
         or print STDERR "* Cannot connect to ftp.gnome.org: $@\n";

  $gnomeftp->login("anonymous",'-anonymous@')
         or print STDERR "* Cannot login ", $gnomeftp->message, "\n";

  my @all_files = glob($prefix . "/$mask");
  foreach my $myfile (@all_files)
  {
    my ($SERVER, $firstpath, $URL, $name, $V, $LATEST) = read_box($myfile);
    ######################################
    # Special handling for sourceforge.net
    ######################################
    if ($SERVER eq 'SF.NET')
    {
      my $NEW_VERSION="?";
      open(CMD,"wget -N http://prdownloads.sourceforge.net/$firstpath/?sort_by=date -O - 2>/dev/null|");
      while (<CMD>)
      {
        chomp;
        my $line = $_;
        if ($line =~ /$URL/)
        {
      if ($1) {
            $NEW_VERSION = "$1";
      } else {
        $NEW_VERSION = "";
      }
        }
      }
      if (($V ne $NEW_VERSION) && ($NEW_VERSION ne '?') && (length($NEW_VERSION) > 0))
      {
        print STDERR "+ SF.NET\n";
        insert_row($dbh, $name, $V, $NEW_VERSION);
      }
      close(CMD);
    } elsif ($SERVER eq 'GNOME.ORG')
    ################################
    # Special handling for gnome.org
    ################################
    {
      my $NEW_VERSION="?";
      my $dir;
      if ($LATEST =~ /ftp\:\/\/ftp\.gnome\.org(.+)/)
      {
        $dir="$1";
      }
      $gnomeftp->cwd("$dir") or print STDERR "* Cannot change working directory ", $gnomeftp->message, "\n";
      my @lines=$gnomeftp->ls(".") or print STDERR "* ls failed: ", $gnomeftp->message, "\n";

      foreach (@lines)
      {
        if (/LATEST-IS-(.+)/)
        {
          $NEW_VERSION="$1";
        }
      }
      if ($V ne $NEW_VERSION && $NEW_VERSION ne '?' && (length($NEW_VERSION) > 0))
      {
        print STDERR "+ GNOME.ORG\n";
        insert_row($dbh, $name, $V, $NEW_VERSION);
      }
    } elsif (($LATEST =~ /ftp\:\/\/([^\/]+)\/(.+)/) || ($SERVER eq 'GNU.ORG'))
    ###################################
    # Special handling for GENERIC ftps
    ###################################
    {
      $LATEST = macro($LATEST, $V, $name);
      $URL =~ s/\%\{GNU\.ORG\}/ftp:\/\/ftp\.gnu\.org\/pub\/gnu/;
      $URL = macro($URL, $V, $name);
      $LATEST =~ /ftp\:\/\/([^\/]+)\/(.+)/;
      my $host = $1;
      my $dir = $2;
      my $ftp;
      if ($ftp = Net::FTP->new("$host", Debug => 0, Timeout => 15, Passive => 1))
      {
        if ($ftp->login("anonymous",'-anonymous@'))
        {
          if ($ftp->cwd("$dir"))
          {
            if (my @lines=$ftp->ls("."))
            {
              my %f;
              $URL =~ s/$/\$/;
              foreach my $file (@lines)
              {
                chomp($file);
                my $mdtm = $ftp->mdtm("$file");
                my $path = "ftp://$host/$dir$file";
                if ($path =~ /$URL/)
                {
                  $f{"$file:$1"} = "$mdtm";
                }
              }
              my $key = (sort { $f{$b} <=> $f{$a} } keys %f)[0];
              my $NEW_VERSION = '?';
          if (defined($key))
          {
            $NEW_VERSION = (split(/\:/, $key))[1];
                if (defined($NEW_VERSION)) {
                if (($V ne $NEW_VERSION) && ($NEW_VERSION ne '?') && (length($NEW_VERSION) > 0))
                {
                  print STDERR "+ GENERIC FTP\n";
                  insert_row($dbh, $name, $V, $NEW_VERSION);
                }
            }
          }
        } else {
           print STDERR "* Dir failed: ", $ftp->message, "\n";
               $ftp->quit;
        }
      } else {
         print STDERR "* Cannot change working directory ", $ftp->message, "\n";
             $ftp->quit;
      }
    } else {
          print STDERR "* Cannot login ", $ftp->message, "\n";
          $ftp->quit;
    }
        $ftp->quit;
      } else {
        print STDERR "* Cannot connect to $host: $@\n";
      }
    } elsif ($LATEST =~ /http\:\/\/(.+)\/([^\/]+)/)
    ###################################
    # Special handling for GENERIC http
    ###################################
    {
    my $dir = "$LATEST";
    $dir = macro($dir, $V, $name);
    my $agent = new LWP::UserAgent;
    $agent->agent("Mozilla/5.0 (X11; U; Linux i686; en-US; rv:1.6) Gecko/20040113)", $agent->agent);
    $agent->timeout(5);

    my $request = new HTTP::Request GET => "$dir/";
    my $response = $agent->request($request);
    my @lines = split(/\n/, $response->content);

    my $url = basename($URL);

    my @maybe;
    foreach my $line (@lines)
    {
      chomp($line);
      if ($line =~ /$url/)
      {
        my $ver = $1;
        if ($line =~ /href\=[\'\"](.+)[\'\"]/i && defined($ver))
        {
	  my $n = $1;
	  my $magic = 0;
	  my $multi = 100000000000000;
          my @numbers = split(/\./, $ver);
	  foreach my $number (@numbers)
	  {
	    $number =~ s/[a-zA-Z\_]//g;
	    if ($number =~ /\-/) {
	      $number = (split(/\-/, $number))[0] + 1;
	    }
	    $magic += $number * $multi;
	    $multi = $multi / 100;
	  }
          push @maybe, "$n|$ver|$magic";
        }
      }
    }
    # pseudo numeric comparision
    my @sorted = sort({ (split(/\|/, $a))[2] <=> (split(/\|/, $b))[2] } @maybe);
    #my @sorted = @maybe;
    if (@sorted)
    {
      my @tab = split(/\|/, $sorted[$#sorted]);
      my $NEW_VERSION = '';
      if (defined($tab[1])) {
        $NEW_VERSION = $tab[1];
      }
          if ($V ne $NEW_VERSION && $NEW_VERSION ne '?' && (length($NEW_VERSION) > 0))
          {
            print STDERR "+ GENERIC HTTP\n";
        insert_row($dbh, $name, $V, $NEW_VERSION);
          }
    }
    }
  }
  # disconnect from server
  $gnomeftp->quit;
  $dbh->disconnect();
}

# ============================================================================
# propose_log()
# ============================================================================

sub propose_log
{
  my ($file) = (@_);
#  print "FIXME: Not implemented yet!\n";

# TODO
# - configure parameters changes!
# - added/removed/commented patch
# - added post-install section (and others)
# - macroized: scrollkeeper-update (temporary)
# - added/removed/changed diffignore
# - added/removed /build/fails (build fail reason)

my @lines = ();
my ($ver1, $ver2, $req1, $req2, $prov1, $prov2, $do1, $do2, $conf1, $conf2);
my ($group1, $group2, $patch1, $patch2, $categories);
my (%options1, %options2, @options3, @options4, @options5, @options6);

open(DIFF,"$file");
@lines = <DIFF>;
close(DIFF);

$group1 = "";
$group2 = "";
$patch1 = "";
$patch2 = "";
$categories = 0;
foreach my $line (@lines) {
  chomp($line);
  if ($line =~ /^\+.*\<patch.*\>(.+)\.(patch|diff)\<\/patch\>/) {
    $patch1 = $1;
  }
  if ($line =~ /^\-.*\<patch.*\>(.+)\.(patch|diff)\<\/patch\>/) {
    $patch2 = $1;
  }
  if ($line =~ /\-.*\<box group\=\"([^\"]+)\"/) {
    $group1 = $1;
  }
  if ($line =~ /\+.*\<box group\=\"([^\"]+)\"/) {
    $group2 = $1;
  }
  if ($line =~ /\-.*\<name version\=\"([^\"]+)\"/) {
    $ver1 = $1;
  }
  if ($line =~ /\+.*\<name version\=\"([^\"]+)\"/) {
    $ver2 = $1;
  }
  if ($line =~ /\-.*\<requires/) {
    $req1 = 1;
  }
  if ($line =~ /\+.*\<requires/) {
    $req2 = 1;
  }
  if ($line =~ /\-.*\<provides/) {
    $prov1 = 1;
  }
  if ($line =~ /\+.*\<provides/) {
    $prov2 = 1;
  }
  if ($line =~ /\-.*\<do/) {
    $do1 = 1;
  }
  if ($line =~ /\+.*\<do/) {
    $do2 = 1;
  }
  if ($line =~ /\-.*\%\{configure\}/) {
    $conf1 = 1;
  }
  if ($line =~ /\+.*\%\{configure_old\}/) {
    $conf2 = 1;
  }
  if ($line =~ /(.).*\-\-(with|without|enable|disable)\-([^\s\=]+)/) {
    if ($1 eq '-') {
      $options1{"$2-$3"} = 1;
    }
    if ($1 eq '+') {
      $options2{"$2-$3"} = 1;
    }
  }
  if ($line =~ /\+.*\%\{categories\}/) {
    $categories = 1;
  }
}

if ($group1 ne $group2) {
  print "- moved to $group2 group\n";
}

if ($ver1 && $ver2) {
  print "- updated to $ver2\n";
}

if ($req1 && $req2) {
  print "- updated requires\n";
}

if ($req1) {
  print "- removed some requires\n";
}

if ($req2 && !$prov2) {
  print "- added some requires\n";
}

if ($prov1 && $prov2) {
  print "- updated provides\n";
}

if ($prov1) {
  print "- removed some provides\n";
}

if ($prov2 && !$req2) {
  print "- added some provides\n";
}

if ($req2 && $prov2) {
  print "- added provides/requires\n";
}

if ($do1 && $do2) {
  print "- updated ToDo\n";
}

if ($do2 && !$do1)  {
  print "- added ToDo item\n";
}

if ($conf1 && $conf2) {
  print "- switched from macro configure to configure_old\n";
}

if (length($patch1) > 0)
{
  print "- added patch $patch1\n";
}

if (length($patch2) > 0)
{
  print "- removed patch $patch2\n";
}

if ($categories == 1)
{
  print "- fixed desktop categories\n";
}

foreach my $key (keys %options1) {
  if (!exists $options2{$key}) {
    push @options3, $key;
  }
}

foreach my $key (keys %options2) {
  if (!exists $options1{$key}) {
    push @options4, $key;
  }
}

foreach my $key (@options4) {
  if ($key =~ /(with|without|enable|disable)\-(.+)/) {
    if ($1 eq 'enable' || $1 eq 'with') {
      push @options5, $2;
    }
    if ($1 eq 'disable' || $1 eq 'without') {
      push @options6, $2;
    }
  }
}

if (@options5) {
  my $ncount = scalar(@options5);
  my $i = 0;
  my $plural='';
  if ($ncount != 1) {
    $plural = 's:';
  }
  print "- enabled option$plural ";
  foreach (@options5) {
    $i++;
    print "$_";
    if ($i < $ncount) {
      print ", ";
    }
  }
  print "\n";
}

if (@options6) {
  my $ncount = scalar(@options6);
  my $i = 0;
  my $plural='';
  if ($ncount != 1) {
    $plural = 's:';
  }
  print "- disabled option$plural ";
  foreach (@options6) {
    $i++;
    print "$_";
    if ($i < $ncount) {
      print ", ";
    }
  }
  print "\n";
}

}

# ============================================================================
# distfiles()
# ============================================================================

sub _distfiles
{
    my ($dbh, $type) = (@_);
    my $age;
    my $attic;
    my @distfiles;
    my @return = ();

    if ($type eq 'attic')
    {
      $age = 2;
      $attic = '';
    } elsif ($type eq 'clean')
    {
      $age = 5;
      $attic = 'Attic/';
    } elsif ($type eq 'invalid')
    {
      @distfiles = glob("$DISTFILES/*");
    }
    if ($type ne 'invalid')
    {
    ######
    my $sth = $dbh->prepare("SELECT DISTINCT program FROM sources ORDER BY program");
    $sth->execute();

    while (my @r = $sth->fetchrow_array)
    {
      my $program = $r[0];

      my $sth2 = $dbh->prepare("select distinct filename from sources where program = '$program' and date >= (select d from (select distinct b, d from (select revision b, filename, max(date) d from sources where program='$program' group by filename order by d DESC) y limit $age,1) z)");
      $sth2->execute();
      my %array1 = ();
      while (my @r2 = $sth2->fetchrow_array)
      {
        $array1{$r2[0]} = 1;
      }
      $sth2->finish();

      my $sth3 = $dbh->prepare("select distinct filename from sources where program = '$program' and date < (select d from (select distinct b, d from (select revision b, filename, max(date) d from sources where program='$program' group by filename order by d DESC) y limit $age,1) z)");
      $sth3->execute();
      my $index = 0;
      my @array2 = ();
      while (my @r3 = $sth3->fetchrow_array)
      {
        $array2[$index] = $r3[0];
        $index++;
      }
      $sth3->finish();

      foreach my $file (@array2)
      {
        if (!exists($array1{$file}))
        {
        if (-f "$DISTFILES/$attic$file")
        {
            push @return, "$file";
        }
        }
      }
    }
    $sth->finish();
    ######
    } else
    {
      # invalid
      my %db_files;

      my $sth4 = $dbh->prepare("SELECT DISTINCT filename FROM sources");
      $sth4->execute();
      while (my @r4 = $sth4->fetchrow_array)
      {
        $db_files{$r4[0]} = 1;
      }
      $sth4->finish();
      foreach my $file (@distfiles)
      {
        $file =~ s/$DISTFILES\///g;
        if (!exists($db_files{$file}))
    {
      if (-f "$DISTFILES/$file")
      {
        push @return, "$file";
      }
    }
      }
    }
    return @return;
}

sub distfiles
{
  use DBI();
  my ($type) = (@_);

  # yes, yes password is here ;)
  my $dbh = DBI->connect("DBI:mysql:database=pingwinek:host=localhost",'root','toor', {'RaiseError' => 1}) or die;

  my @return = _distfiles($dbh, 'invalid');
  foreach my $r (@return) {
    print "rm -f '$r'\n";
  } 

  @return = _distfiles($dbh, 'attic');
  foreach my $r (@return) {
    print "mv '$r' Attic\n";
  }

  @return = _distfiles($dbh, 'clean');
  foreach my $r (@return) {
    print "rm -f 'Attic/$r'\n";
  }

  $dbh->disconnect();
}

# ============================================================================
# cvs2db()
# ============================================================================

sub _check_new_version
{
  my ($dbh, $prefix, $filename, %CVS_revisions) = (@_);
  my $log = 0;
  my $revision;
  my $date;
  my $V;
  my $P;
  my @temp;
  my $stop = 0;

  @temp = xpath("$prefix/$filename",'/box/source/name');
  my $HEAD_name = getfirst(@temp);
  my $sth = $dbh->prepare("SELECT revision FROM sources WHERE program='$HEAD_name' ORDER BY date DESC LIMIT 1");
  $sth->execute();
  my $r = $sth->fetchrow_hashref();
  my $DB_revision = $r->{'revision'};
  $sth->finish();

  if (!defined($DB_revision)) {
    $DB_revision = '';
  }

  if ($DB_revision ne $CVS_revisions{$filename})
  {
    open(FILE,"cd $prefix; cvs log $filename|");
    while (<FILE>)
    {
    chomp;
    my $line = $_;
    if ($line =~ /^description\:$/) {
      $log = 1;
    }
    if (($log == 1) && ($stop == 0))
    {
      if ($line =~ /revision ([0-9\.]+)/) {
        $revision = $1;
        if ($revision eq $DB_revision) {
          $stop = 1;
        }
      }
      if ($stop == 0)
      {
      if ($line =~ /^date: ([0-9]{4}\/[0-9]{2}\/[0-9]{2} [0-9]{2}\:[0-5][0-9]\:[0-5][0-9])\;/)
      {
        $date = $1;
        sleep(1);
        system("cd $prefix; cvs update -A -r $revision $filename &>/dev/null");

        @temp = xpath("$prefix/$filename",'/box/source/name/@version');
        $V = getfirst(@temp);
        @temp = xpath("$prefix/$filename",'/box/source/name');
        $P = getfirst(@temp);
        @temp = xpath("$prefix/$filename",'/box/source/url');
        my @URLS=();
        foreach my $url (@temp)
        {
          $url = macro($url, $V, $P);
          if (length($url) > 0) {
            push @URLS, $url;
          }
        }
        foreach my $url (@URLS)
        {
            $url = basename($url);
            $dbh->do("INSERT INTO sources (program, version, filename, revision, date) VALUES ('$P', '$V', '$url', '$revision', '$date')");
            print STDERR "INSERTED: $P\t$V\t$revision\n";
        }
        $revision = '';
        $date = '';
      }
    }
  }
}
}
close(FILE);
}


sub cvs2db
{
  use DBI();

  my ($prefix) = (@_);

  # check if we are in correct directory
  my $cvsroot = ':pserver:anonymous@cvs.gna.org/var/cvs/pingwinek/';
  my $cvsroot2 = 'robot@cvs.gna.org:/var/cvs/pingwinek';
  # if not defined use default prefix
  if (!defined($prefix))
  {
    $prefix = $BOXES . '-anonymous';
  }
  if (!((-f "$prefix/CVS/Root") && (-f "$prefix/CVS/Repository")))
  {
    print STDERR "You should be in checkouted CVS boxes repository!\n";
    return;
  } else
  {
    my $valid_repo = 0;
    my $valid_root = 0;
    open(CVS, "$prefix/CVS/Repository");
    while (<CVS>)
    {
      if (/^boxes$/) {
        $valid_repo = 1;
      }
    }
    close(CVS);

    open(CVS, "$prefix/CVS/Root");
    while (<CVS>)
    {
      chomp;
      my $line = $_;
      if (($line eq "$cvsroot") || ($line eq "$cvsroot2")) {
        $valid_root = 1;
      }
    }
    close(CVS);

    if ($valid_root == 0)
    {
      print STDERR "Invalid CVSROOT! Should be: $cvsroot\n";
      return;
    } elsif ($valid_repo == 0)
    {
      print STDERR "Invalid repository! Should be: boxes\n";
      return;
    }
  }

  # yes, yes password is here ;)
  my $dbh = DBI->connect("DBI:mysql:database=pingwinek:host=localhost",'root','toor', {'RaiseError' => 1}) or die;
  # update boxes from HEAD
  system("cd $prefix; cvs update -dP -A -rHEAD 2>&1 >/dev/null");
  my %CVS_revisions;

  open(CVS,"$prefix/CVS/Entries");
  while (<CVS>)
  {
    if (/^\//)
    {
      my ($x, $f, $r) = split(/\//);
      $CVS_revisions{$f} = $r;
    }
  }
  close(CVS);

  foreach my $filename (glob("$prefix/*.box"))
  {
    $filename =~ s/$prefix\///;
    print STDERR "Processing $filename...\n";
    _check_new_version($dbh, $prefix, $filename, %CVS_revisions);
  }
  $dbh->disconnect();
}

# ============================================================================
# prepare_patch($src_box)
# ============================================================================

sub prepare_patch
{
  my ($file) = (@_);
  my (@temp, $P, $V, $R, $URL, $COMPRESS, $root, @patches);
  my $PATCH_TMP = "$TEMPDIR/PATCHES";
    
  my $FILE = "package.box";
  open(CMD,"tar xf $file $FILE 2>/dev/null|");
  close(CMD);
  if (! -f $FILE) {
    die "Bad file!\n";
  }
  @temp = xpath("$FILE",'/box/source[@id="main"]/name/@version');
  $V = getfirst(@temp);
  @temp = xpath("$FILE",'/box/source[@id="main"]/name/@release');
  $R = getfirst(@temp);
  @temp = xpath("$FILE",'/box/source[@id="main"]/name');
  $P = getfirst(@temp);
  @temp = xpath("$FILE",'/box/source[@id="main"]/url');
  $URL = macro(basename(getfirst(@temp)), $V);
  @temp = xpath("$FILE",'/box/source[@id="main"]/@root');
  $root =getfirst(@temp);
  @patches = xpath("$FILE",'/box/source[@id="main"]/patch');
  
  if (($URL =~ /tar\.gz/) || ($URL =~ /tgz/)) {
    $COMPRESS = "z";
  }
  if (($URL =~ /tar\.bz2/) || ($URL =~ /tbz/)) {
    $COMPRESS = "j";
  }
  
  system("rm -rf $PATCH_TMP/$P-$V");
  system("mkdir -p $PATCH_TMP/$P-$V");
  system("tar xfO $file sources/$URL | tar xvf$COMPRESS - -C $PATCH_TMP/$P-$V --overwrite");

  my $dir = '';
  if (length($root) > 0)
  {
    if ($root !~ /^\./)
    {
      $dir = $root;
    }
  } else {
    my @dirs = glob("$PATCH_TMP/$P-$V/*");
    if (@dirs == 1) {
      if (-d $dirs[0]) {
        $dir = basename($dirs[0]);
      }
    }
  }
  if (length($dir) > 0)
  {
    my $patch_files = '';
    foreach (@patches) {
      $patch_files .= "patches/$_ ";
    }
    print "$patch_files\n";
    system("cd $PATCH_TMP/$P-$V; cp -r $dir $dir.orig");
    # -C in tar don't work :(
    
    chomp(my $cwd = `pwd`);
    my $newfile = "$cwd/" . basename($file);
    system("cd $PATCH_TMP/$P-$V; tar xvf $newfile $patch_files  --overwrite ");
    system("cd $PATCH_TMP/$P-$V; cp patches/* . 2>/dev/null; mv patches patches-orig;");
    
    # apply helper
    open(CMDFILE, ">$PATCH_TMP/$P-$V/apply");
    print CMDFILE q{#!/bin/sh

[ $# != 1 ] && exit
};
    print CMDFILE "cd $P-$V\n";
    print CMDFILE q{
patch -p1 < ../$1    
};
    close(CMDFILE);
  }
  system("chmod 755 $PATCH_TMP/$P-$V/apply");
  
  # create helper
  open(CMDFILE, ">$PATCH_TMP/$P-$V/create");
  print CMDFILE q{#!/bin/sh

[ $# != 1 ] && exit
};
  print CMDFILE "diff -urN $P-$V.orig $P-$V > \$1\n";
  close(CMDFILE);
  system("chmod 755 $PATCH_TMP/$P-$V/create");

  # reset helper
  open(CMDFILE, ">$PATCH_TMP/$P-$V/reset");
  print CMDFILE q{#!/bin/sh
};
  print CMDFILE "rm -rf $P-$V\n";
  print CMDFILE "cp -r $P-$V.orig $P-$V\n";
  close(CMDFILE);
  system("chmod 755 $PATCH_TMP/$P-$V/reset");

  # deploy helper
  open(CMDFILE, ">$PATCH_TMP/$P-$V/deploy");
  print CMDFILE q{#!/bin/sh

cp *.patch *.diff /pingwinek/CVS/patches/
};
  close(CMDFILE);
  system("chmod 755 $PATCH_TMP/$P-$V/deploy");

  system("rm -f $FILE");
}

# ============================================================================
# bsdiff($bin1, $bin2)
# ============================================================================

sub bsdiff
{
  my ($f1,$f2) = (@_);
  open(F,">boxdiff.sh");
  print F "B1=$f1\n";
  print F "B2=$f2\n";
  print F q{

mkdir -p dir1 dir2
cd dir1
tar xf ../$B1
d=$(ls -l data | cut -d' ' -f7,8)
gunzip data/*
touch --date "$d" data
cd ..

cd dir2
tar xf ../$B2
d=$(ls -l data | cut -d' ' -f7,8)
gunzip data/*
touch --date "$d" data
cd ..

cd dir1
tar cf ../dir1.tar *
cd ..

cd dir2
tar cf ../dir2.tar *
cd ..

name=${B2%box}
bsdiff dir1.tar dir2.tar ${name}pbox
rm -rf dir1 dir2 dir1.tar dir2.tar
rm -f boxdiff.sh
};
  close(F);
  system "sh boxdiff.sh";
}

# ============================================================================
# bspatch($bin1, $bin2)
# ============================================================================

sub bspatch
{
  my ($f1,$f2) = (@_);
  open(F,">boxpatch.sh");
  print F "B1=$f1\n";
  print F "B3=$f2\n";
  print F q{
B2=${B3%pbox}box

mkdir -p apply
cd apply
# 1a
tar xf ../$B1
# 2a
d=$(ls -l data | cut -d' ' -f7,8)
gunzip data/*
touch --date "$d" data
cd ..


cd apply
# 3a
tar cf ../apply.tar *
cd ..

# 4
bspatch apply.tar new.tar $B3

cd apply
rm -rf *
tar xf ../new.tar
gzip data/*
rm ../new.tar

tar cf ../$B2 *
cd ..

rm -rf apply apply.tar
rm -f boxpatch.sh
};
  close(F);
  system "sh boxpatch.sh";
}

# ============================================================================
# show_file($file, $box)
# ============================================================================

sub show_file
{
  my ($show, $file) = (@_);
  my $type = "";
  my @array = split(/\./, $file);
  my $last = $#array;
  my @files = ();
  if ($last < 2) {
    $type = "";
  } else {
    $type = $array[$#array - 1];
  }
  if (-f $file)
  {
    if (($array[$last] eq "box") and ($type ne "src")) {
      my $code = system "tar xf $file package.box &>/dev/null";
      if ($code == 0)
      {
        my @temp = xpath("package.box",'/box/source/name');
#        print STDERR getfirst(@temp) . "\n";
        $file =~ /(.*)\.box$/;
	my $datafile = "$1.tar.gz";
	my @arr = split(/\//, $datafile);
	$datafile = $arr[$#arr];
	system "rm -f package.box";
	open(CACHE,"tar xfO $file cache/$show 2>/dev/null|");
	@files = <CACHE>;
	close(CACHE);
	return @files;
      } else
      {
        print "File $file is not binary package!\n";
	return ();
      }
    } else
    {
      print "File $file is not binary package!\n";
      return ();
    }
  } else
  {
    print "File $file does not exists!\n";
    return ();
  }
}

# ============================================================================
# show_deps($package)
# ============================================================================

sub show_deps
{
  my ($type, $package) = (@_);

  use DBI;
  my $dbh;
  my $sth;
  my $sql_type;
  
  if ($type eq "install") {
    $sql_type = "= 0";
  } elsif ($type eq "build") {
    $sql_type = "= 1";
  } elsif ($type eq "all") {
    $sql_type = ">= 0";
  } 
  
  $dbh = DBI->connect("DBI:SQLite:dbname=$BOXROOT/FACTORY/latest/database",'','', {'RaiseError' => 1}) or die;
  $sth = $dbh->prepare("select name from packages where id in (select distinct pr.package_id from
  requires r, packages p, provides pr where p.id = r.package_id and
  p.name = '$package' and r.type $sql_type and pr.name = r.name) and name != '$package' and grp >=0
  order by name;");
  $sth->execute();
  while (my @row = $sth->fetchrow_array)
  {
    printf "%s\n", $row[0];
  }  
}

sub topology
{
  use DBI;
  my $dbh;
  my $sth;
  my $sth2;
  my $sql_type;

  $dbh = DBI->connect("DBI:SQLite:dbname=$BOXROOT/FACTORY/latest/database",'','', {'RaiseError' => 1}) or die;
  
  # only for core packages now
  $sth = $dbh->prepare("select name from packages where grp >= 0");
  $sth->execute();
  while (my @row = $sth->fetchrow_array)
  {
    my $package = $row[0];
  $sth2 = $dbh->prepare("select name from packages where id in (select distinct pr.package_id from
  requires r, packages p, provides pr where p.id = r.package_id and
  p.name = '$package' and r.type = 0 and pr.name = r.name) and name != '$package'
  order by name");
  $sth2->execute();
  my $was=0;
  while (my @row2 = $sth2->fetchrow_array)
  {
    printf "%s\t%s\n", $row2[0],$package;
    $was=1;
  }  
  if ($was == 0)
  {
    printf "%s\t%s\n", $package, $package;
  }
  }  
  
}


# ============================================================================
# what_provide($what)
# ============================================================================

sub what_provide
{
  my ($what, $ver) = (@_);
  if (!defined($ver)) {
    $ver = '';
  }
  use DBI;
  my $dbh;
  my $sth;
  
  $dbh = DBI->connect("DBI:SQLite:dbname=$BOXROOT/FACTORY/latest/database",'','', {'RaiseError' => 1}) or die;
  $sth = $dbh->prepare("select pkg.name from packages pkg, provides p where 
  pkg.id = p.package_id and p.name = '$what'");
  $sth->execute();
  while (my @row = $sth->fetchrow_array)
  {
    printf "%s\t%s\n", $row[0], $ver;
  }  
}


# ============================================================================
# Main()
# ============================================================================

my $params = $#ARGV;
my $file;
my $tag;
my $force;
my $debug;

if ($params==-1)
{
  usage();
} elsif ($params>=0)
{
  read_config;
  my $option = $ARGV[0];
  if ($option eq '-s' and $params>=1)
  {
    # FIXME: add support for -d (debug)
    $file = $ARGV[$params];
    $tag = '0';
    $debug = '0';
    if ($params == 3) {
      $tag = $ARGV[2];
    }
    generate_source_package $file, $tag, $debug;
  } elsif ($option eq '-b' and $params>=1)
  {
    $file = $ARGV[1];
    build_package($file);
  } elsif ($option eq '-l' and $params>=1)
  {
    $file = $ARGV[1];
    foreach (list_package($file)) {
      chomp;
      print "$_\n";
    }
  } elsif ($option eq '-i' and $params>=1)
  {
    $file = $ARGV[$params];
    my $root = "/";
    if ($params == 3)
    {
      if ($ARGV[1] eq '-root')
      {
        $root = $ARGV[2];
      }
    }
    install_package($file, $root);
  } elsif ($option eq '-g' and $params>=1)
  {
    $file = $ARGV[$params];
    $force = '0';
    $tag = '0';
    if ($params >= 2)
    {
      for (my $i=1; $i<$params; $i++)
      {
        if ($ARGV[$i] eq '-f') {
	  $force = 1;
	}
	if ($ARGV[$i] eq '-r' and $i < $params - 1) {
          $tag = $ARGV[$i + 1];
	}
      }
    }
    download_source $file, $force, $tag;
  } elsif ($option eq '-ga' and $params>=0)
  {
    $force = '0';
    $tag = '0';
    if ($params >= 1)
    {
      for (my $i=1; $i<=$params; $i++)
      {
        if ($ARGV[$i] eq '-f') {
	  $force = 1;
	}
	if ($ARGV[$i] eq '-r' and $i < $params) {
          $tag = $ARGV[$i + 1];
	}
      }
    }
    download_all_sources $force, $tag;
  } elsif ($option eq '-todo' and $params>=0)
  {
    $tag = '0';
    if ($params >= 2)
    {
      for (my $i=1; $i<=$params; $i++)
      {
	if ($ARGV[$i] eq '-r' and $i < $params) {
          $tag = $ARGV[$i + 1];
	}
      }
    }
    generate_todo_list $tag;
  } elsif ($option eq '-diff' and $params>=2)
  {
    my $file1 = $ARGV[1];
    my $file2 = $ARGV[2];
    diff_package($file1, $file2);
  } elsif ($option eq '-validate' and $params>=1)
  {
    $file = $ARGV[1];
    my $exitcode = validate_package($file);
    if ($exitcode == 1) {
	print STDERR "Error validating file $file!\n";
    }
    exit $exitcode;
  } elsif ($option eq '-init')
  {
    initialize();
  } elsif ($option eq '-template')
  {
    my $url = $ARGV[1];
    template($url);
  } elsif ($option eq '-check' and $params>=1)
  {
    $file = $ARGV[1];
    check_factory($file);
  } elsif ($option eq '-accept' and $params>=1)
  {
    $file = $ARGV[1];
    accept_factory($file);
  } elsif (($option eq '-checkdep' ||
            $option eq '-checkbuilddep' ||
	    $option eq '-checkfiles') and $params>=0)
  {
    my @files = ();
    if ($params == 0) {
      # default group when no specified
      push @files, 'core';
      push @files, 'extras';
    } else {
      for (my $i = 1; $i <= $#ARGV; $i++) {
        push @files, $ARGV[$i];
      }
    }
    if ($option eq '-checkdep') {
      check_dependency("install",@files);
    } elsif ($option eq '-checkbuilddep') {
      check_dependency("build",@files);
    } elsif ($option eq '-checkfiles') {
      check_files(@files);
    }
  } elsif ($option eq '-tidy')
  {
    tidy_box();
  } elsif ($option eq '-report')
  {
    generate_report();
  } elsif ($option eq '-newsoft')
  {
    my $prefix = $ARGV[1];
    print_newsoft($prefix);
  } elsif ($option eq '-proposelog' and $params>=1)
  {
    $file = $ARGV[1];
    propose_log($file);
  } elsif ($option eq '-distfiles') 
  {
    distfiles('?');
  } elsif ($option eq '-cvs2db')
  {
    my $prefix = $ARGV[1];
    cvs2db($prefix);
  } elsif ($option eq '-prepatch' and $params>=1)
  {
    $file = $ARGV[1];
    prepare_patch($file);
  } elsif ($option eq '-bsdiff' and $params==2)
  {
    my $file1 = $ARGV[1];
    my $file2 = $ARGV[2];
    bsdiff($file1, $file2);
  } elsif ($option eq '-bspatch' and $params==2)
  {
    my $file1 = $ARGV[1];
    my $file2 = $ARGV[2];
    bspatch($file1, $file2);
  } elsif ($option eq '-show' and $params==2)
  {
    my $file1 = $ARGV[1];
    my $file2 = $ARGV[2];
    foreach (show_file($file1, $file2)) {
      chomp;
      print "$_\n";
    }
  } elsif ($option eq '-deps' and $params==1)
  {
    show_deps("all", $ARGV[1]);
  } elsif ($option eq '-deps_build' and $params==1)
  {
    show_deps("build", $ARGV[1]);
  } elsif ($option eq '-deps_install' and $params==1)
  {
    show_deps("install", $ARGV[1]);
  } elsif ($option eq '-what_provide' and $params>=1)
  {
    what_provide($ARGV[1], $ARGV[2]);
  } elsif ($option eq '-topology')
  {
    topology();
  } else
  {
    usage();
  }
}

exit 0;
